from __future__ import annotations

from typing import Literal, Optional

import numpy as np
import pyqtgraph as pg
from bec_lib.logger import bec_logger
from pydantic import Field, ValidationError, field_validator
from qtpy.QtCore import Signal

from bec_widgets.utils import BECConnector, Colors, ConnectionConfig
from bec_widgets.widgets.plots.image.image_processor import (
    ImageProcessor,
    ImageStats,
    ProcessingConfig,
)

logger = bec_logger.logger


# noinspection PyDataclass
class ImageItemConfig(ConnectionConfig):  # TODO review config
    parent_id: str | None = Field(None, description="The parent plot of the image.")
    monitor: str | None = Field(None, description="The name of the monitor.")
    monitor_type: Literal["1d", "2d", "auto"] = Field("auto", description="The type of monitor.")
    source: str | None = Field(None, description="The source of the curve.")
    color_map: str | None = Field("plasma", description="The color map of the image.")
    downsample: bool | None = Field(True, description="Whether to downsample the image.")
    opacity: float | None = Field(1.0, description="The opacity of the image.")
    v_range: tuple[float | int, float | int] | None = Field(
        None, description="The range of the color bar. If None, the range is automatically set."
    )
    autorange: bool | None = Field(True, description="Whether to autorange the color bar.")
    autorange_mode: Literal["max", "mean"] = Field(
        "mean", description="Whether to use the mean of the image for autoscaling."
    )
    processing: ProcessingConfig = Field(
        default_factory=ProcessingConfig, description="The post processing of the image."
    )

    model_config: dict = {"validate_assignment": True}
    _validate_color_map = field_validator("color_map")(Colors.validate_color_map)


class ImageItem(BECConnector, pg.ImageItem):
    RPC = True
    USER_ACCESS = [
        "color_map",
        "color_map.setter",
        "v_range",
        "v_range.setter",
        "v_min",
        "v_min.setter",
        "v_max",
        "v_max.setter",
        "autorange",
        "autorange.setter",
        "autorange_mode",
        "autorange_mode.setter",
        "fft",
        "fft.setter",
        "log",
        "log.setter",
        "num_rotation_90",
        "num_rotation_90.setter",
        "transpose",
        "transpose.setter",
        "get_data",
    ]

    vRangeChangedManually = Signal(tuple)

    def __init__(
        self,
        config: Optional[ImageItemConfig] = None,
        gui_id: Optional[str] = None,
        parent_image=None,
        **kwargs,
    ):
        if config is None:
            config = ImageItemConfig(widget_class=self.__class__.__name__)
            self.config = config
        else:
            self.config = config
        if parent_image is not None:
            self.set_parent(parent_image)
        else:
            self.parent_image = None
        super().__init__(config=config, gui_id=gui_id, **kwargs)

        self.raw_data = None
        self.buffer = []
        self.max_len = 0

        # Image processor will handle any setting of data
        self._image_processor = ImageProcessor(config=self.config.processing)

    def set_parent(self, parent: BECConnector):
        self.parent_image = parent

    def parent(self):
        return self.parent_image

    def set_data(self, data: np.ndarray):
        self.raw_data = data
        self._process_image()

    ################################################################################
    # Properties
    @property
    def color_map(self) -> str:
        """Get the current color map."""
        return self.config.color_map

    @color_map.setter
    def color_map(self, value: str):
        """Set a new color map."""
        try:
            self.config.color_map = value
            self.setColorMap(value)
        except ValidationError:
            logger.error(f"Invalid colormap '{value}' provided.")

    @property
    def v_range(self) -> tuple[float, float]:
        """
        Get the color intensity range of the image.
        """
        if self.levels is not None:
            return tuple(float(x) for x in self.levels)
        return 0.0, 1.0

    @v_range.setter
    def v_range(self, vrange: tuple[float, float]):
        """
        Set the color intensity range of the image.
        """
        self.set_v_range(vrange, disable_autorange=True)

    def set_v_range(self, vrange: tuple[float, float], disable_autorange=True):
        if disable_autorange:
            self.config.autorange = False
            self.vRangeChangedManually.emit(vrange)
        self.setLevels(vrange)
        self.config.v_range = vrange

    @property
    def v_min(self) -> float:
        return self.v_range[0]

    @v_min.setter
    def v_min(self, value: float):
        self.v_range = (value, self.v_range[1])

    @property
    def v_max(self) -> float:
        return self.v_range[1]

    @v_max.setter
    def v_max(self, value: float):
        self.v_range = (self.v_range[0], value)

    ################################################################################
    # Autorange Logic

    @property
    def autorange(self) -> bool:
        return self.config.autorange

    @autorange.setter
    def autorange(self, value: bool):
        self.config.autorange = value
        if value:
            self.apply_autorange()

    @property
    def autorange_mode(self) -> str:
        return self.config.autorange_mode

    @autorange_mode.setter
    def autorange_mode(self, mode: str):
        self.config.autorange_mode = mode
        if self.autorange:
            self.apply_autorange()

    def apply_autorange(self):
        if self.raw_data is None:
            return
        data = self.image
        if data is None:
            data = self.raw_data
        stats = ImageStats.from_data(data)
        self.auto_update_vrange(stats)

    def auto_update_vrange(self, stats: ImageStats) -> None:
        """Update the v_range based on the stats of the image."""
        fumble_factor = 2
        if self.config.autorange_mode == "mean":
            vmin = max(stats.mean - fumble_factor * stats.std, 0)
            vmax = stats.mean + fumble_factor * stats.std
        elif self.config.autorange_mode == "max":
            vmin, vmax = stats.minimum, stats.maximum
        else:
            return
        self.set_v_range(vrange=(vmin, vmax), disable_autorange=False)

    ################################################################################
    # Data Processing Logic

    def _process_image(self):
        """
        Reprocess the current raw data and update the image display.
        """
        if self.raw_data is not None:
            autorange = self.config.autorange
            self._image_processor.set_config(self.config.processing)
            processed_data = self._image_processor.process_image(self.raw_data)
            self.setImage(processed_data, autoLevels=False)
            self.autorange = autorange

    @property
    def fft(self) -> bool:
        """Get or set whether FFT postprocessing is enabled."""
        return self.config.processing.fft

    @fft.setter
    def fft(self, enable: bool):
        self.config.processing.fft = enable
        self._process_image()

    @property
    def log(self) -> bool:
        """Get or set whether logarithmic scaling is applied."""
        return self.config.processing.log

    @log.setter
    def log(self, enable: bool):
        self.config.processing.log = enable
        self._process_image()

    @property
    def num_rotation_90(self) -> Optional[int]:
        """Get or set the number of 90° rotations to apply."""
        return self.config.processing.num_rotation_90

    @num_rotation_90.setter
    def num_rotation_90(self, value: Optional[int]):
        self.config.processing.num_rotation_90 = value
        self._process_image()

    @property
    def transpose(self) -> bool:
        """Get or set whether the image is transposed."""
        return self.config.processing.transpose

    @transpose.setter
    def transpose(self, enable: bool):
        self.config.processing.transpose = enable
        self._process_image()

    ################################################################################
    # Export
    def get_data(self) -> np.ndarray:
        """
        Get the data of the image.
        Returns:
            np.ndarray: The data of the image.
        """
        return self.image

    def clear(self):
        super().clear()
        self.raw_data = None
        self.buffer = []
        self.max_len = 0

    def remove(self):
        self.parent().disconnect_monitor(self.config.monitor)
        self.clear()



from __future__ import annotations

from collections import defaultdict
from typing import Any

import numpy as np
import pyqtgraph as pg
from qtpy.QtCore import QObject, Qt, Signal, Slot
from qtpy.QtWidgets import QApplication


class CrosshairScatterItem(pg.ScatterPlotItem):
    def setDownsampling(self, ds=None, auto=None, method=None):
        pass

    def setClipToView(self, state):
        pass

    def setAlpha(self, *args, **kwargs):
        pass


class Crosshair(QObject):
    # QT Position of mouse cursor
    positionChanged = Signal(tuple)
    positionClicked = Signal(tuple)
    # Plain crosshair position signals mapped to real coordinates
    crosshairChanged = Signal(tuple)
    crosshairClicked = Signal(tuple)
    # Signal for 1D plot
    coordinatesChanged1D = Signal(tuple)
    coordinatesClicked1D = Signal(tuple)
    # Signal for 2D plot
    coordinatesChanged2D = Signal(tuple)
    coordinatesClicked2D = Signal(tuple)

    def __init__(
        self,
        plot_item: pg.PlotItem,
        precision: int | None = None,
        *,
        min_precision: int = 2,
        parent=None,
    ):
        """
        Crosshair for 1D and 2D plots.

        Args:
            plot_item (pyqtgraph.PlotItem): The plot item to which the crosshair will be attached.
            precision (int | None, optional): Fixed number of decimal places to display. If *None*, precision is chosen dynamically from the current view range.
            min_precision (int, optional): The lower bound (in decimal places) used when dynamic precision is enabled. Defaults to 2.
            parent (QObject, optional): Parent object for the QObject. Defaults to None.
        """
        super().__init__(parent)
        self.is_log_y = None
        self.is_log_x = None
        self.is_derivative = None
        self.plot_item = plot_item
        self._precision = precision
        self._min_precision = max(0, int(min_precision))  # ensure non‑negative

        self.v_line = pg.InfiniteLine(angle=90, movable=False)
        self.v_line.skip_auto_range = True
        self.h_line = pg.InfiniteLine(angle=0, movable=False)
        self.h_line.skip_auto_range = True
        # Add custom attribute to identify crosshair lines
        self.v_line.is_crosshair = True
        self.h_line.is_crosshair = True
        self.plot_item.addItem(self.v_line, ignoreBounds=True)
        self.plot_item.addItem(self.h_line, ignoreBounds=True)

        # Initialize highlighted curve in a case of multiple curves
        self.highlighted_curve_index = None

        # Add TextItem to display coordinates
        self.coord_label = pg.TextItem("", anchor=(1, 1), fill=(0, 0, 0, 100))
        self.coord_label.setVisible(False)  # Hide initially
        self.coord_label.skip_auto_range = True
        self.plot_item.addItem(self.coord_label)

        # Signals to connect
        self.proxy = pg.SignalProxy(
            self.plot_item.scene().sigMouseMoved, rateLimit=60, slot=self.mouse_moved
        )
        self.positionChanged.connect(self.update_coord_label)
        self.plot_item.scene().sigMouseClicked.connect(self.mouse_clicked)

        # Connect signals from pyqtgraph right click menu
        self.plot_item.ctrl.derivativeCheck.checkStateChanged.connect(self.check_derivatives)
        self.plot_item.ctrl.logXCheck.checkStateChanged.connect(self.check_log)
        self.plot_item.ctrl.logYCheck.checkStateChanged.connect(self.check_log)
        self.plot_item.ctrl.downsampleSpin.valueChanged.connect(self.clear_markers)

        # Initialize markers
        self.items = []
        self.marker_moved_1d = {}
        self.marker_clicked_1d = {}
        self.marker_2d_row = None
        self.marker_2d_col = None
        self.update_markers()
        self.check_log()
        self.check_derivatives()

        self._connect_to_theme_change()

    @property
    def precision(self) -> int | None:
        """Fixed number of decimals; ``None`` enables dynamic mode."""
        return self._precision

    @precision.setter
    def precision(self, value: int | None):
        """
        Set the fixed number of decimals to display.

        Args:
            value(int | None): The number of decimals to display. If `None`, dynamic precision is used based on the view range.
        """
        self._precision = value

    @property
    def min_precision(self) -> int:
        """Lower bound on decimals when dynamic precision is used."""
        return self._min_precision

    @min_precision.setter
    def min_precision(self, value: int):
        """
        Set the lower bound on decimals when dynamic precision is used.

        Args:
            value(int): The minimum number of decimals to display. Must be non-negative.
        """
        self._min_precision = max(0, int(value))

    def _current_precision(self) -> int:
        """
        Get the current precision based on the view range or fixed precision.
        """
        if self._precision is not None:
            return self._precision

        # Dynamically choose precision from the smaller visible span
        view_range = self.plot_item.vb.viewRange()
        x_span = abs(view_range[0][1] - view_range[0][0])
        y_span = abs(view_range[1][1] - view_range[1][0])

        # Ignore zero spans that can appear during initialisation
        spans = [s for s in (x_span, y_span) if s > 0]
        span = min(spans) if spans else 1.0

        exponent = np.floor(np.log10(span))  # order of magnitude
        decimals = max(0, int(-exponent) + 1)
        return max(self._min_precision, decimals)

    def _connect_to_theme_change(self):
        """Connect to the theme change signal."""
        qapp = QApplication.instance()
        if hasattr(qapp, "theme_signal"):
            qapp.theme_signal.theme_updated.connect(self._update_theme)
            self._update_theme()

    @Slot(str)
    def _update_theme(self, theme: str | None = None):
        """Update the theme."""
        if theme is None:
            qapp = QApplication.instance()
            if hasattr(qapp, "theme"):
                theme = qapp.theme.theme
            else:
                theme = "dark"
        self.apply_theme(theme)

    def apply_theme(self, theme: str):
        """Apply the theme to the plot."""
        if theme == "dark":
            text_color = "w"
            label_bg_color = (50, 50, 50, 150)
        elif theme == "light":
            text_color = "k"
            label_bg_color = (240, 240, 240, 150)
        else:
            text_color = "w"
            label_bg_color = (50, 50, 50, 150)

        self.coord_label.setColor(text_color)
        self.coord_label.fill = pg.mkBrush(label_bg_color)
        self.coord_label.border = pg.mkPen(None)

    @Slot(int)
    def update_highlighted_curve(self, curve_index: int):
        """
        Update the highlighted curve in the case of multiple curves in a plot item.

        Args:
            curve_index(int): The index of curve to highlight
        """
        self.highlighted_curve_index = curve_index
        self.clear_markers()
        self.update_markers()

    def update_markers(self):
        """Update the markers for the crosshair, creating new ones if necessary."""

        if self.highlighted_curve_index is not None and hasattr(self.plot_item, "visible_curves"):
            # Focus on the highlighted curve only
            self.items = [self.plot_item.visible_curves[self.highlighted_curve_index]]
        else:
            # Handle all curves
            self.items = self.plot_item.items

        # Create or update markers
        for item in self.items:
            if isinstance(item, pg.PlotDataItem):  # 1D plot
                pen = item.opts["pen"]
                color = pen.color() if hasattr(pen, "color") else pg.mkColor(pen)
                name = item.name() or str(id(item))
                if name in self.marker_moved_1d:
                    # Update existing markers
                    marker_moved = self.marker_moved_1d[name]
                    marker_moved.setPen(pg.mkPen(color))
                    # Update clicked markers' brushes
                    for marker_clicked in self.marker_clicked_1d[name]:
                        alpha = marker_clicked.opts["brush"].color().alpha()
                        marker_clicked.setBrush(
                            pg.mkBrush(color.red(), color.green(), color.blue(), alpha)
                        )
                    # Update z-values
                    marker_moved.setZValue(item.zValue() + 1)
                    for marker_clicked in self.marker_clicked_1d[name]:
                        marker_clicked.setZValue(item.zValue() + 1)
                else:
                    # Create new markers
                    marker_moved = CrosshairScatterItem(
                        size=10, pen=pg.mkPen(color), brush=pg.mkBrush(None)
                    )
                    marker_moved.skip_auto_range = True
                    marker_moved.is_crosshair = True
                    self.marker_moved_1d[name] = marker_moved
                    self.plot_item.addItem(marker_moved)
                    # Set marker z-value higher than the curve
                    marker_moved.setZValue(item.zValue() + 1)

                    # Create glowing effect markers for clicked events
                    marker_clicked_list = []
                    for size, alpha in [(18, 64), (14, 128), (10, 255)]:
                        marker_clicked = CrosshairScatterItem(
                            size=size,
                            pen=pg.mkPen(None),
                            brush=pg.mkBrush(color.red(), color.green(), color.blue(), alpha),
                        )
                        marker_clicked.skip_auto_range = True
                        marker_clicked.is_crosshair = True
                        self.plot_item.addItem(marker_clicked)
                        marker_clicked.setZValue(item.zValue() + 1)
                        marker_clicked_list.append(marker_clicked)
                    self.marker_clicked_1d[name] = marker_clicked_list
            elif isinstance(item, pg.ImageItem):  # 2D plot
                if self.marker_2d_row is not None and self.marker_2d_col is not None:
                    continue
                # Create horizontal ROI for row highlighting
                if item.image is None:
                    continue
                self.marker_2d_row = pg.ROI(
                    [0, 0], size=[item.image.shape[0], 1], pen=pg.mkPen("r", width=2), movable=False
                )
                self.marker_2d_row.skip_auto_range = True
                self.plot_item.addItem(self.marker_2d_row)

                # Create vertical ROI for column highlighting
                self.marker_2d_col = pg.ROI(
                    [0, 0], size=[1, item.image.shape[1]], pen=pg.mkPen("r", width=2), movable=False
                )
                self.marker_2d_col.skip_auto_range = True
                self.plot_item.addItem(self.marker_2d_col)

    def snap_to_data(
        self, x: float, y: float
    ) -> tuple[None, None] | tuple[defaultdict[Any, list], defaultdict[Any, list]]:
        """
        Finds the nearest data points to the given x and y coordinates.

        Args:
            x(float): The x-coordinate of the mouse cursor
            y(float): The y-coordinate of the mouse cursor

        Returns:
            tuple: x and y values snapped to the nearest data
        """
        y_values = defaultdict(list)
        x_values = defaultdict(list)

        # Iterate through items in the plot
        for item in self.items:
            if isinstance(item, pg.PlotDataItem):  # 1D plot
                name = item.name() or str(id(item))
                plot_data = item._getDisplayDataset()
                if plot_data is None:
                    continue
                x_data, y_data = plot_data.x, plot_data.y
                if x_data is not None and y_data is not None:
                    if self.is_log_x:
                        min_x_data = np.min(x_data[x_data > 0])
                    else:
                        min_x_data = np.min(x_data)
                    max_x_data = np.max(x_data)
                    if x < min_x_data or x > max_x_data:
                        y_values[name] = None
                        x_values[name] = None
                        continue
                    closest_x, closest_y = self.closest_x_y_value(x, x_data, y_data)
                    y_values[name] = closest_y
                    x_values[name] = closest_x
            elif isinstance(item, pg.ImageItem):  # 2D plot
                name = item.config.monitor or str(id(item))
                image_2d = item.image
                if image_2d is None:
                    continue
                # Clip the x and y values to the image dimensions to avoid out of bounds errors
                y_values[name] = int(np.clip(y, 0, image_2d.shape[1] - 1))
                x_values[name] = int(np.clip(x, 0, image_2d.shape[0] - 1))

        if x_values and y_values:
            if all(v is None for v in x_values.values()) or all(
                v is None for v in y_values.values()
            ):
                return None, None
            return x_values, y_values

        return None, None

    def closest_x_y_value(self, input_x: float, list_x: list, list_y: list) -> tuple:
        """
        Find the closest x and y value to the input value.

        Args:
            input_x (float): Input value
            list_x (list): List of x values
            list_y (list): List of y values

        Returns:
            tuple: Closest x and y value
        """
        # Convert lists to NumPy arrays
        arr_x = np.asarray(list_x)

        # Get the indices where x is not NaN
        valid_indices = ~np.isnan(arr_x)

        # Filter x array to exclude NaN values
        filtered_x = arr_x[valid_indices]

        # Find the index of the closest value in the filtered x array
        closest_index = np.abs(filtered_x - input_x).argmin()

        # Map back to the original index in the list_x and list_y arrays
        original_index = np.where(valid_indices)[0][closest_index]

        return list_x[original_index], list_y[original_index]

    def mouse_moved(self, event):
        """Handles the mouse moved event, updating the crosshair position and emitting signals.

        Args:
            event: The mouse moved event
        """
        pos = event[0]
        self.update_markers()
        if self.plot_item.vb.sceneBoundingRect().contains(pos):
            mouse_point = self.plot_item.vb.mapSceneToView(pos)
            x, y = mouse_point.x(), mouse_point.y()
            self.v_line.setPos(x)
            self.h_line.setPos(y)
            scaled_x, scaled_y = self.scale_emitted_coordinates(mouse_point.x(), mouse_point.y())
            self.crosshairChanged.emit((scaled_x, scaled_y))
            self.positionChanged.emit((x, y))

            x_snap_values, y_snap_values = self.snap_to_data(x, y)
            if x_snap_values is None or y_snap_values is None:
                return
            if all(v is None for v in x_snap_values.values()) or all(
                v is None for v in y_snap_values.values()
            ):
                # not sure how we got here, but just to be safe...
                return

            precision = self._current_precision()
            for item in self.items:
                if isinstance(item, pg.PlotDataItem):
                    name = item.name() or str(id(item))
                    x, y = x_snap_values[name], y_snap_values[name]
                    if x is None or y is None:
                        continue
                    self.marker_moved_1d[name].setData([x], [y])
                    x_snapped_scaled, y_snapped_scaled = self.scale_emitted_coordinates(x, y)
                    coordinate_to_emit = (
                        name,
                        round(x_snapped_scaled, precision),
                        round(y_snapped_scaled, precision),
                    )
                    self.coordinatesChanged1D.emit(coordinate_to_emit)
                elif isinstance(item, pg.ImageItem):
                    name = item.config.monitor or str(id(item))
                    x, y = x_snap_values[name], y_snap_values[name]
                    if x is None or y is None:
                        continue
                    # Set position of horizontal ROI (row)
                    self.marker_2d_row.setPos([0, y])
                    # Set position of vertical ROI (column)
                    self.marker_2d_col.setPos([x, 0])
                    coordinate_to_emit = (name, x, y)
                    self.coordinatesChanged2D.emit(coordinate_to_emit)
                else:
                    continue

    def mouse_clicked(self, event):
        """Handles the mouse clicked event, updating the crosshair position and emitting signals.

        Args:
            event: The mouse clicked event
        """

        # we only accept left mouse clicks
        if event.button() != Qt.MouseButton.LeftButton:
            return
        self.update_markers()
        if self.plot_item.vb.sceneBoundingRect().contains(event._scenePos):
            mouse_point = self.plot_item.vb.mapSceneToView(event._scenePos)
            x, y = mouse_point.x(), mouse_point.y()
            scaled_x, scaled_y = self.scale_emitted_coordinates(mouse_point.x(), mouse_point.y())
            self.crosshairClicked.emit((scaled_x, scaled_y))
            self.positionClicked.emit((x, y))

            x_snap_values, y_snap_values = self.snap_to_data(x, y)

            if x_snap_values is None or y_snap_values is None:
                return
            if all(v is None for v in x_snap_values.values()) or all(
                v is None for v in y_snap_values.values()
            ):
                # not sure how we got here, but just to be safe...
                return

            precision = self._current_precision()
            for item in self.items:
                if isinstance(item, pg.PlotDataItem):
                    name = item.name() or str(id(item))
                    x, y = x_snap_values[name], y_snap_values[name]
                    if x is None or y is None:
                        continue
                    for marker_clicked in self.marker_clicked_1d[name]:
                        marker_clicked.setData([x], [y])
                    x_snapped_scaled, y_snapped_scaled = self.scale_emitted_coordinates(x, y)
                    coordinate_to_emit = (
                        name,
                        round(x_snapped_scaled, precision),
                        round(y_snapped_scaled, precision),
                    )
                    self.coordinatesClicked1D.emit(coordinate_to_emit)
                elif isinstance(item, pg.ImageItem):
                    name = item.config.monitor or str(id(item))
                    x, y = x_snap_values[name], y_snap_values[name]
                    if x is None or y is None:
                        continue
                    # Set position of horizontal ROI (row)
                    self.marker_2d_row.setPos([0, y])
                    # Set position of vertical ROI (column)
                    self.marker_2d_col.setPos([x, 0])
                    coordinate_to_emit = (name, x, y)
                    self.coordinatesClicked2D.emit(coordinate_to_emit)
                else:
                    continue

    def clear_markers(self):
        """Clears the markers from the plot."""
        for marker in self.marker_moved_1d.values():
            self.plot_item.removeItem(marker)
        for markers in self.marker_clicked_1d.values():
            for marker in markers:
                self.plot_item.removeItem(marker)
        self.marker_moved_1d.clear()
        self.marker_clicked_1d.clear()

    def scale_emitted_coordinates(self, x, y):
        """Scales the emitted coordinates if the axes are in log scale.

        Args:
            x (float): The x-coordinate
            y (float): The y-coordinate

        Returns:
            tuple: The scaled x and y coordinates
        """
        if self.is_log_x:
            x = 10**x
        if self.is_log_y:
            y = 10**y
        return x, y

    def update_coord_label(self, pos: tuple):
        """Updates the coordinate label based on the crosshair position and axis scales.

        Args:
            pos (tuple): The (x, y) position of the crosshair.
        """
        x, y = pos
        x_scaled, y_scaled = self.scale_emitted_coordinates(x, y)
        precision = self._current_precision()
        text = f"({x_scaled:.{precision}f}, {y_scaled:.{precision}f})"
        for item in self.items:
            if isinstance(item, pg.ImageItem):
                image = item.image
                if image is None:
                    continue
                ix = int(np.clip(x, 0, image.shape[0] - 1))
                iy = int(np.clip(y, 0, image.shape[1] - 1))
                intensity = image[ix, iy]
                text += f"\nIntensity: {intensity:.{precision}f}"
                break
        # Update coordinate label
        self.coord_label.setText(text)
        self.coord_label.setPos(x, y)
        self.coord_label.setVisible(True)

    def check_log(self):
        """Checks if the x or y axis is in log scale and updates the internal state accordingly."""
        self.is_log_x = self.plot_item.axes["bottom"]["item"].logMode
        self.is_log_y = self.plot_item.axes["left"]["item"].logMode
        self.clear_markers()

    def check_derivatives(self):
        """Checks if the derivatives are enabled and updates the internal state accordingly."""
        self.is_derivative = self.plot_item.ctrl.derivativeCheck.isChecked()
        self.clear_markers()

    def cleanup(self):
        if self.marker_2d_row is not None:
            self.plot_item.removeItem(self.marker_2d_row)
            self.marker_2d_row = None
        if self.marker_2d_col is not None:
            self.plot_item.removeItem(self.marker_2d_col)
            self.marker_2d_col = None
        self.plot_item.removeItem(self.v_line)
        self.plot_item.removeItem(self.h_line)
        self.plot_item.removeItem(self.coord_label)

        self.clear_markers()



import os
import select
import shlex
import signal
import socket
import subprocess
from typing import Literal

from pydantic import BaseModel
from qtpy.QtCore import Signal, Slot

from bec_widgets.widgets.editors.website.website import WebsiteWidget


class VSCodeInstructionMessage(BaseModel):
    command: Literal["open", "write", "close", "zenMode", "save", "new", "setCursor"]
    content: str = ""


def get_free_port():
    """
    Get a free port on the local machine.

    Returns:
        int: The free port number
    """
    sock = socket.socket()
    sock.bind(("", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


class VSCodeEditor(WebsiteWidget):
    """
    A widget to display the VSCode editor.
    """

    file_saved = Signal(str)

    token = "bec"
    host = "127.0.0.1"

    PLUGIN = True
    USER_ACCESS = []
    ICON_NAME = "developer_mode_tv"

    def __init__(self, parent=None, config=None, client=None, gui_id=None, **kwargs):

        self.process = None
        self.port = get_free_port()
        self._url = f"http://{self.host}:{self.port}?tkn={self.token}"
        super().__init__(parent=parent, config=config, client=client, gui_id=gui_id, **kwargs)
        self.start_server()
        self.bec_dispatcher.connect_slot(self.on_vscode_event, f"vscode-events/{self.gui_id}")

    def start_server(self):
        """
        Start the server.

        This method starts the server for the VSCode editor in a subprocess.
        """

        env = os.environ.copy()
        env["BEC_Widgets_GUIID"] = self.gui_id
        env["BEC_REDIS_HOST"] = self.client.connector.host
        cmd = shlex.split(
            f"code serve-web --port {self.port} --connection-token={self.token} --accept-server-license-terms"
        )
        self.process = subprocess.Popen(
            cmd,
            text=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            preexec_fn=os.setsid,
            env=env,
        )

        os.set_blocking(self.process.stdout.fileno(), False)
        while self.process.poll() is None:
            readylist, _, _ = select.select([self.process.stdout], [], [], 1)
            if self.process.stdout in readylist:
                output = self.process.stdout.read(1024)
                if output and f"available at {self._url}" in output:
                    break
        self.set_url(self._url)
        self.wait_until_loaded()

    @Slot(str)
    def open_file(self, file_path: str):
        """
        Open a file in the VSCode editor.

        Args:
            file_path: The file path to open
        """
        msg = VSCodeInstructionMessage(command="open", content=f"file://{file_path}")
        self.client.connector.raw_send(f"vscode-instructions/{self.gui_id}", msg.model_dump_json())

    @Slot(dict, dict)
    def on_vscode_event(self, content, _metadata):
        """
        Handle the VSCode event. VSCode events are received as RawMessages.

        Args:
            content: The content of the event
            metadata: The metadata of the event
        """

        # the message also contains the content but I think is fine for now to just emit the file path
        if not isinstance(content["data"], dict):
            return
        if "uri" not in content["data"]:
            return
        if not content["data"]["uri"].startswith("file://"):
            return
        file_path = content["data"]["uri"].split("file://")[1]
        self.file_saved.emit(file_path)

    @Slot()
    def save_file(self):
        """
        Save the file in the VSCode editor.
        """
        msg = VSCodeInstructionMessage(command="save")
        self.client.connector.raw_send(f"vscode-instructions/{self.gui_id}", msg.model_dump_json())

    @Slot()
    def new_file(self):
        """
        Create a new file in the VSCode editor.
        """
        msg = VSCodeInstructionMessage(command="new")
        self.client.connector.raw_send(f"vscode-instructions/{self.gui_id}", msg.model_dump_json())

    @Slot()
    def close_file(self):
        """
        Close the file in the VSCode editor.
        """
        msg = VSCodeInstructionMessage(command="close")
        self.client.connector.raw_send(f"vscode-instructions/{self.gui_id}", msg.model_dump_json())

    @Slot(str)
    def write_file(self, content: str):
        """
        Write content to the file in the VSCode editor.

        Args:
            content: The content to write
        """
        msg = VSCodeInstructionMessage(command="write", content=content)
        self.client.connector.raw_send(f"vscode-instructions/{self.gui_id}", msg.model_dump_json())

    @Slot()
    def zen_mode(self):
        """
        Toggle the Zen mode in the VSCode editor.
        """
        msg = VSCodeInstructionMessage(command="zenMode")
        self.client.connector.raw_send(f"vscode-instructions/{self.gui_id}", msg.model_dump_json())

    @Slot(int, int)
    def set_cursor(self, line: int, column: int):
        """
        Set the cursor in the VSCode editor.

        Args:
            line: The line number
            column: The column number
        """
        msg = VSCodeInstructionMessage(command="setCursor", content=f"{line},{column}")
        self.client.connector.raw_send(f"vscode-instructions/{self.gui_id}", msg.model_dump_json())

    def cleanup_vscode(self):
        """
        Cleanup the VSCode editor.
        """
        if not self.process or self.process.poll() is not None:
            return
        os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
        self.process.wait()

    def cleanup(self):
        """
        Cleanup the widget. This method is called from the dock area when the widget is removed.
        """
        self.bec_dispatcher.disconnect_slot(self.on_vscode_event, f"vscode-events/{self.gui_id}")
        self.cleanup_vscode()
        return super().cleanup()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    widget = VSCodeEditor(gui_id="unknown")
    widget.show()
    app.exec_()
    widget.bec_dispatcher.disconnect_all()
    widget.client.shutdown()



# Copyright (C) 2022 The Qt Company Ltd.
# SPDX-License-Identifier: LicenseRef-Qt-Commercial OR BSD-3-Clause

"""PySide6 port of the Qt Designer taskmenuextension example from Qt v6.x"""

import sys

from bec_ipython_client.main import BECIPythonClient
from qtpy.QtWidgets import QApplication

from bec_widgets.examples.plugin_example_pyside.tictactoe import TicTacToe

if __name__ == "__main__":  # pragma: no cover
    app = QApplication(sys.argv)
    window = TicTacToe()
    window.state = "-X-XO----"
    window.show()
    sys.exit(app.exec())



def main():  # pragma: no cover
    from qtpy import PYSIDE6

    if not PYSIDE6:
        print("PYSIDE6 is not available in the environment. Cannot patch designer.")
        return
    from PySide6.QtDesigner import QPyDesignerCustomWidgetCollection

    from bec_widgets.widgets.editors.web_console.web_console_plugin import WebConsolePlugin

    QPyDesignerCustomWidgetCollection.addCustomWidget(WebConsolePlugin())


if __name__ == "__main__":  # pragma: no cover
    main()



from __future__ import annotations

import secrets
import subprocess
import time

from bec_lib.logger import bec_logger
from louie.saferef import safe_ref
from qtpy.QtCore import QUrl, qInstallMessageHandler
from qtpy.QtWebEngineWidgets import QWebEnginePage, QWebEngineView
from qtpy.QtWidgets import QApplication, QVBoxLayout, QWidget

from bec_widgets.utils.bec_widget import BECWidget

logger = bec_logger.logger


class WebConsoleRegistry:
    """
    A registry for the WebConsole class to manage its instances.
    """

    def __init__(self):
        """
        Initialize the registry.
        """
        self._instances = {}
        self._server_process = None
        self._server_port = None
        self._token = secrets.token_hex(16)

    def register(self, instance: WebConsole):
        """
        Register an instance of WebConsole.
        """
        self._instances[instance.gui_id] = safe_ref(instance)
        self.cleanup()

        if self._server_process is None:
            # Start the ttyd server if not already running
            self.start_ttyd()

    def start_ttyd(self, use_zsh: bool | None = None):
        """
        Start the ttyd server
        ttyd -q -W -t 'theme={"background": "black"}' zsh

        Args:
            use_zsh (bool): Whether to use zsh or bash. If None, it will try to detect if zsh is available.
        """

        # First, check if ttyd is installed
        try:
            subprocess.run(["ttyd", "--version"], check=True, stdout=subprocess.PIPE)
        except FileNotFoundError:
            # pylint: disable=raise-missing-from
            raise RuntimeError("ttyd is not installed. Please install it first.")

        if use_zsh is None:
            # Check if we can use zsh
            try:
                subprocess.run(["zsh", "--version"], check=True, stdout=subprocess.PIPE)
                use_zsh = True
            except FileNotFoundError:
                use_zsh = False

        command = [
            "ttyd",
            "-p",
            "0",
            "-W",
            "-t",
            'theme={"background": "black"}',
            "-c",
            f"user:{self._token}",
        ]
        if use_zsh:
            command.append("zsh")
        else:
            command.append("bash")

        # Start the ttyd server
        self._server_process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )

        self._wait_for_server_port()

        self._server_process.stdout.close()
        self._server_process.stderr.close()

    def _wait_for_server_port(self, timeout: float = 10):
        """
        Wait for the ttyd server to start and get the port number.

        Args:
            timeout (float): The timeout in seconds to wait for the server to start.
        """
        start_time = time.time()
        while True:
            output = self._server_process.stderr.readline()
            if output == b"" and self._server_process.poll() is not None:
                break
            if not output:
                continue

            output = output.decode("utf-8").strip()
            if "Listening on" in output:
                # Extract the port number from the output
                self._server_port = int(output.split(":")[-1])
                logger.info(f"ttyd server started on port {self._server_port}")
                break
            if time.time() - start_time > timeout:
                raise TimeoutError(
                    "Timeout waiting for ttyd server to start. Please check if ttyd is installed and available in your PATH."
                )

    def cleanup(self):
        """
        Clean up the registry by removing any instances that are no longer valid.
        """
        for gui_id, weak_ref in list(self._instances.items()):
            if weak_ref() is None:
                del self._instances[gui_id]

        if not self._instances and self._server_process:
            # If no instances are left, terminate the server process
            self._server_process.terminate()
            self._server_process = None
            self._server_port = None
            logger.info("ttyd server terminated")

    def unregister(self, instance: WebConsole):
        """
        Unregister an instance of WebConsole.

        Args:
            instance (WebConsole): The instance to unregister.
        """
        if instance.gui_id in self._instances:
            del self._instances[instance.gui_id]

        self.cleanup()


_web_console_registry = WebConsoleRegistry()


def suppress_qt_messages(type_, context, msg):
    if context.category in ["js", "default"]:
        return
    print(msg)


qInstallMessageHandler(suppress_qt_messages)


class BECWebEnginePage(QWebEnginePage):
    def javaScriptConsoleMessage(self, level, message, lineNumber, sourceID):
        logger.info(f"[JS Console] {level.name} at line {lineNumber} in {sourceID}: {message}")


class WebConsole(BECWidget, QWidget):
    """
    A simple widget to display a website
    """

    PLUGIN = True
    ICON_NAME = "terminal"

    def __init__(self, parent=None, config=None, client=None, gui_id=None, **kwargs):
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)
        _web_console_registry.register(self)
        self._token = _web_console_registry._token
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.browser = QWebEngineView(self)
        self.page = BECWebEnginePage(self)
        self.page.authenticationRequired.connect(self._authenticate)
        self.browser.setPage(self.page)
        layout.addWidget(self.browser)
        self.setLayout(layout)
        self.page.setUrl(QUrl(f"http://localhost:{_web_console_registry._server_port}"))

    def write(self, data: str, send_return: bool = True):
        """
        Send data to the web page
        """
        self.page.runJavaScript(f"window.term.paste('{data}');")
        if send_return:
            self.send_return()

    def _authenticate(self, _, auth):
        """
        Authenticate the request with the provided username and password.
        """
        auth.setUser("user")
        auth.setPassword(self._token)

    def send_return(self):
        """
        Send return to the web page
        """
        self.page.runJavaScript(
            "document.querySelector('textarea.xterm-helper-textarea').dispatchEvent(new KeyboardEvent('keypress', {charCode: 13}))"
        )

    def send_ctrl_c(self):
        """
        Send Ctrl+C to the web page
        """
        self.page.runJavaScript(
            "document.querySelector('textarea.xterm-helper-textarea').dispatchEvent(new KeyboardEvent('keypress', {charCode: 3}))"
        )

    def cleanup(self):
        """
        Clean up the registry by removing any instances that are no longer valid.
        """
        _web_console_registry.unregister(self)
        super().cleanup()


if __name__ == "__main__":  # pragma: no cover
    import sys

    app = QApplication(sys.argv)
    widget = WebConsole()
    widget.show()
    sys.exit(app.exec_())



from __future__ import annotations

from typing import TYPE_CHECKING

from bec_lib.logger import bec_logger
from qtpy.QtCore import QMimeData, Qt
from qtpy.QtGui import QDrag
from qtpy.QtWidgets import QApplication, QHBoxLayout, QLabel, QWidget

if TYPE_CHECKING:  # pragma: no cover
    from qtpy.QtGui import QMouseEvent

logger = bec_logger.logger


class DeviceItem(QWidget):
    def __init__(self, device: str) -> None:
        super().__init__()

        self._drag_pos = None

        self.device = device
        layout = QHBoxLayout()
        layout.setContentsMargins(10, 2, 10, 2)
        self.label = QLabel(device)
        layout.addWidget(self.label)
        self.setLayout(layout)

        self.setStyleSheet(
            """
            border: 1px solid #ddd;
            border-radius: 5px;
            padding: 10px;
        """
        )

    def mousePressEvent(self, event: QMouseEvent) -> None:
        super().mousePressEvent(event)
        if event.button() == Qt.LeftButton:
            self._drag_pos = event.pos()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        if not (event.buttons() and Qt.LeftButton):
            return
        if (event.pos() - self._drag_pos).manhattanLength() < QApplication.startDragDistance():
            return

        drag = QDrag(self)
        mime_data = QMimeData()
        mime_data.setText(self.device)
        drag.setMimeData(mime_data)
        drag.exec_(Qt.MoveAction)

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        logger.debug("Double Clicked")
        # TODO: Implement double click action for opening the device properties dialog
        return super().mouseDoubleClickEvent(event)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    widget = DeviceItem("Device")
    widget.show()
    sys.exit(app.exec_())



from __future__ import annotations

from abc import abstractmethod
from decimal import Decimal
from types import UnionType
from typing import Callable, Protocol

from bec_lib.logger import bec_logger
from bec_qthemes import material_icon
from pydantic import BaseModel, ConfigDict, Field
from pydantic.fields import FieldInfo
from qtpy.QtCore import Signal  # type: ignore
from qtpy.QtWidgets import (
    QApplication,
    QButtonGroup,
    QCheckBox,
    QDoubleSpinBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QLineEdit,
    QRadioButton,
    QSpinBox,
    QToolButton,
    QWidget,
)

from bec_widgets.widgets.editors.scan_metadata._util import (
    clearable_required,
    field_default,
    field_limits,
    field_maxlen,
    field_minlen,
    field_precision,
)

logger = bec_logger.logger


class FormItemSpec(BaseModel):
    """
    The specification for an item in a dynamically generated form. Uses a pydantic FieldInfo
    to store most annotation info, since one of the main purposes is to store data for
    forms genrated from pydantic models, but can also be composed from other sources or by hand.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    item_type: type | UnionType
    name: str
    info: FieldInfo = FieldInfo()


class ClearableBoolEntry(QWidget):
    stateChanged = Signal()

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._layout = QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)
        self._layout.setSizeConstraint(QLayout.SizeConstraint.SetFixedSize)
        self._entry = QButtonGroup()
        self._true = QRadioButton("true", parent=self)
        self._false = QRadioButton("false", parent=self)
        for button in [self._true, self._false]:
            self._layout.addWidget(button)
            self._entry.addButton(button)
            button.toggled.connect(self.stateChanged)

    def clear(self):
        self._entry.setExclusive(False)
        self._true.setChecked(False)
        self._false.setChecked(False)
        self._entry.setExclusive(True)

    def isChecked(self) -> bool | None:
        if not self._true.isChecked() and not self._false.isChecked():
            return None
        return self._true.isChecked()

    def setChecked(self, value: bool | None):
        if value is None:
            self.clear()
        elif value:
            self._true.setChecked(True)
            self._false.setChecked(False)
        else:
            self._true.setChecked(False)
            self._false.setChecked(True)

    def setToolTip(self, tooltip: str):
        self._true.setToolTip(tooltip)
        self._false.setToolTip(tooltip)


class DynamicFormItem(QWidget):
    valueChanged = Signal()

    def __init__(self, parent: QWidget | None = None, *, spec: FormItemSpec) -> None:
        super().__init__(parent)
        self._spec = spec
        self._layout = QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSizeConstraint(QLayout.SizeConstraint.SetMaximumSize)
        self._default = field_default(self._spec.info)
        self._desc = self._spec.info.description
        self.setLayout(self._layout)
        self._add_main_widget()
        if clearable_required(spec.info):
            self._add_clear_button()

    @abstractmethod
    def getValue(self): ...

    @abstractmethod
    def setValue(self, value): ...

    @abstractmethod
    def _add_main_widget(self) -> None:
        """Add the main data entry widget to self._main_widget and appply any
        constraints from the field info"""

    def _describe(self, pad=" "):
        return pad + (self._desc if self._desc else "")

    def _add_clear_button(self):
        self._clear_button = QToolButton()
        self._clear_button.setIcon(
            material_icon(icon_name="close", size=(10, 10), convert_to_pixmap=False)
        )
        self._layout.addWidget(self._clear_button)
        # the widget added in _add_main_widget must implement .clear() if value is not required
        self._clear_button.setToolTip("Clear value or reset to default.")
        self._clear_button.clicked.connect(self._main_widget.clear)  # type: ignore

    def _value_changed(self, *_, **__):
        self.valueChanged.emit()


class StrMetadataField(DynamicFormItem):
    def __init__(self, parent: QWidget | None = None, *, spec: FormItemSpec) -> None:
        super().__init__(parent=parent, spec=spec)
        self._main_widget.textChanged.connect(self._value_changed)

    def _add_main_widget(self) -> None:
        self._main_widget = QLineEdit()
        self._layout.addWidget(self._main_widget)
        min_length, max_length = (field_minlen(self._spec.info), field_maxlen(self._spec.info))
        if max_length:
            self._main_widget.setMaxLength(max_length)
        self._main_widget.setToolTip(
            f"(length min: {min_length} max: {max_length}){self._describe()}"
        )
        if self._default:
            self._main_widget.setText(self._default)
            self._add_clear_button()

    def getValue(self):
        if self._main_widget.text() == "":
            return self._default
        return self._main_widget.text()

    def setValue(self, value: str):
        if value is None:
            self._main_widget.setText("")
        self._main_widget.setText(value)


class IntMetadataField(DynamicFormItem):
    def __init__(self, parent: QWidget | None = None, *, spec: FormItemSpec) -> None:
        super().__init__(parent=parent, spec=spec)
        self._main_widget.textChanged.connect(self._value_changed)

    def _add_main_widget(self) -> None:
        self._main_widget = QSpinBox()
        self._layout.addWidget(self._main_widget)
        min_, max_ = field_limits(self._spec.info, int)
        self._main_widget.setMinimum(min_)
        self._main_widget.setMaximum(max_)
        self._main_widget.setToolTip(f"(range {min_} to {max_}){self._describe()}")
        if self._default is not None:
            self._main_widget.setValue(self._default)
            self._add_clear_button()
        else:
            self._main_widget.clear()

    def getValue(self):
        if self._main_widget.text() == "":
            return self._default
        return self._main_widget.value()

    def setValue(self, value: int):
        if value is None:
            self._main_widget.clear()
        self._main_widget.setValue(value)


class FloatDecimalMetadataField(DynamicFormItem):
    def __init__(self, parent: QWidget | None = None, *, spec: FormItemSpec) -> None:
        super().__init__(parent=parent, spec=spec)
        self._main_widget.textChanged.connect(self._value_changed)

    def _add_main_widget(self) -> None:
        self._main_widget = QDoubleSpinBox()
        self._layout.addWidget(self._main_widget)
        min_, max_ = field_limits(self._spec.info, int)
        self._main_widget.setMinimum(min_)
        self._main_widget.setMaximum(max_)
        precision = field_precision(self._spec.info)
        if precision:
            self._main_widget.setDecimals(precision)
        minstr = f"{float(min_):.3f}" if abs(min_) <= 1000 else f"{float(min_):.3e}"
        maxstr = f"{float(max_):.3f}" if abs(max_) <= 1000 else f"{float(max_):.3e}"
        self._main_widget.setToolTip(f"(range {minstr} to {maxstr}){self._describe()}")
        if self._default is not None:
            self._main_widget.setValue(self._default)
            self._add_clear_button()
        else:
            self._main_widget.clear()

    def getValue(self):
        if self._main_widget.text() == "":
            return self._default
        return self._main_widget.value()

    def setValue(self, value: float):
        if value is None:
            self._main_widget.clear()
        self._main_widget.setValue(value)


class BoolMetadataField(DynamicFormItem):
    def __init__(self, *, parent: QWidget | None = None, spec: FormItemSpec) -> None:
        super().__init__(parent=parent, spec=spec)
        self._main_widget.stateChanged.connect(self._value_changed)

    def _add_main_widget(self) -> None:
        if clearable_required(self._spec.info):
            self._main_widget = ClearableBoolEntry()
        else:
            self._main_widget = QCheckBox()
        self._layout.addWidget(self._main_widget)
        self._main_widget.setToolTip(self._describe(""))
        self._main_widget.setChecked(self._default)  # type: ignore # if there is no default then it will be ClearableBoolEntry and can be set with None

    def getValue(self):
        return self._main_widget.isChecked()

    def setValue(self, value):
        self._main_widget.setChecked(value)


def widget_from_type(annotation: type | UnionType | None) -> type[DynamicFormItem]:
    if annotation in [str, str | None]:
        return StrMetadataField
    if annotation in [int, int | None]:
        return IntMetadataField
    if annotation in [float, float | None, Decimal, Decimal | None]:
        return FloatDecimalMetadataField
    if annotation in [bool, bool | None]:
        return BoolMetadataField
    else:
        logger.warning(f"Type {annotation} is not (yet) supported in metadata form creation.")
        return StrMetadataField


if __name__ == "__main__":  # pragma: no cover

    class TestModel(BaseModel):
        value1: str | None = Field(None)
        value2: bool | None = Field(None)
        value3: bool = Field(True)
        value4: int = Field(123)
        value5: int | None = Field()

    app = QApplication([])
    w = QWidget()
    layout = QGridLayout()
    w.setLayout(layout)
    for i, (field_name, info) in enumerate(TestModel.model_fields.items()):
        layout.addWidget(QLabel(field_name), i, 0)
        layout.addWidget(widget_from_type(info.annotation)(info), i, 1)

    w.show()
    app.exec()



from __future__ import annotations

import sys
import traceback
from typing import TYPE_CHECKING

from bec_lib.device import Device, Signal
from bec_lib.endpoints import MessageEndpoints
from bec_qthemes import material_icon
from qtpy.QtCore import Signal as QSignal
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.bec_connector import ConnectionConfig
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import get_accent_colors
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.utils.ophyd_kind_util import Kind
from bec_widgets.widgets.control.device_input.base_classes.device_input_base import (
    DeviceInputConfig,
)
from bec_widgets.widgets.control.device_input.base_classes.device_signal_input_base import (
    DeviceSignalInputBaseConfig,
)
from bec_widgets.widgets.control.device_input.device_line_edit.device_line_edit import (
    DeviceLineEdit,
)
from bec_widgets.widgets.control.device_input.signal_combobox.signal_combobox import SignalComboBox

if TYPE_CHECKING:
    from bec_lib.client import BECClient


class ChoiceDialog(QDialog):
    accepted_output = QSignal(str, str)

    CONNECTION_ERROR_STR = "Error: client is not connected!"

    def __init__(
        self,
        parent: QWidget | None = None,
        config: ConnectionConfig | None = None,
        client: BECClient | None = None,
        show_hinted: bool = True,
        show_normal: bool = False,
        show_config: bool = False,
    ):
        if not client or not client.started:
            self._display_error()
            return
        super().__init__(parent=parent)
        self.setWindowTitle("Choose device and signal...")
        self._accent_colors = get_accent_colors()

        layout = QHBoxLayout()

        config_dict = config.model_dump() if config is not None else {}
        self._device_config = DeviceInputConfig.model_validate(config_dict)
        self._signal_config = DeviceSignalInputBaseConfig.model_validate(config_dict)
        self._device_field = DeviceLineEdit(
            config=self._device_config, parent=parent, client=client
        )
        self._signal_field = SignalComboBox(
            config=self._signal_config,
            device=self._signal_config.device,
            parent=parent,
            client=client,
        )
        layout.addWidget(self._device_field)
        layout.addWidget(self._signal_field)

        self.button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        layout.addWidget(self.button_box)

        self._signal_field.include_hinted_signals = show_hinted
        self._signal_field.include_normal_signals = show_normal
        self._signal_field.include_config_signals = show_config

        self.setLayout(layout)
        self._device_field.textChanged.connect(self._update_device)
        self._device_field.setText(config.device if config is not None else "")

    def _display_error(self):
        try:
            super().__init__()
        except Exception:
            ...
        layout = QHBoxLayout()
        layout.addWidget(QLabel(self.CONNECTION_ERROR_STR))
        self.button_box = QDialogButtonBox(QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.reject)
        layout.addWidget(self.button_box)
        self.setLayout(layout)

    @SafeSlot(str)
    def _update_device(self, device: str):
        if device in self._device_field.dev:
            self._device_field.set_device(device)
            self._signal_field.set_device(device)
            self._device_field.setStyleSheet(
                f"QLineEdit {{ border-style: solid; border-width: 2px; border-color: {self._accent_colors.success.name() if self._accent_colors else 'green'}}}"
            )
            self.button_box.button(QDialogButtonBox.Ok).setEnabled(True)
        else:
            self._device_field.setStyleSheet(
                f"QLineEdit {{ border-style: solid; border-width: 2px; border-color: {self._accent_colors.emergency.name() if self._accent_colors else 'red'}}}"
            )
            self.button_box.button(QDialogButtonBox.Ok).setEnabled(False)
            self._signal_field.clear()

    def accept(self):
        self.accepted_output.emit(self._device_field.text(), self._signal_field.currentText())
        return super().accept()


class SignalLabel(BECWidget, QWidget):

    ICON_NAME = "scoreboard"
    RPC = True
    PLUGIN = True

    USER_ACCESS = [
        "custom_label",
        "custom_units",
        "custom_label.setter",
        "custom_units.setter",
        "decimal_places",
        "decimal_places.setter",
        "show_default_units",
        "show_default_units.setter",
        "show_select_button",
        "show_select_button.setter",
    ]

    def __init__(
        self,
        parent: QWidget | None = None,
        client: BECClient | None = None,
        device: str | None = None,
        signal: str | None = None,
        show_select_button: bool = True,
        show_default_units: bool = False,
        custom_label: str = "",
        custom_units: str = "",
        **kwargs,
    ):
        """Initialize the SignalLabel widget.

        Args:
            parent (QWidget, optional): The parent widget. Defaults to None.
            client (BECClient, optional): The BEC client. Defaults to None.
            device (str, optional): The device name. Defaults to None.
            signal (str, optional): The signal name. Defaults to None.
            selection_dialog_config (DeviceSignalInputBaseConfig | dict, optional): Configuration for the signal selection dialog.
            show_select_button (bool, optional): Whether to show the select button. Defaults to True.
            show_default_units (bool, optional): Whether to show default units. Defaults to False.
            custom_label (str, optional): Custom label for the widget. Defaults to "".
            custom_units (str, optional): Custom units for the widget. Defaults to "".
        """
        self._config = DeviceSignalInputBaseConfig(default=signal, device=device)
        super().__init__(parent=parent, client=client, **kwargs)

        self._device = device
        self._signal = signal

        self._custom_label: str = custom_label
        self._custom_units: str = custom_units
        self._show_default_units: bool = show_default_units
        self._decimal_places = 3

        self._show_hinted_signals: bool = True
        self._show_normal_signals: bool = False
        self._show_config_signals: bool = False

        self._outer_layout = QHBoxLayout()
        self._layout = QHBoxLayout()
        self._outer_layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._outer_layout)

        self._label = QGroupBox(custom_label)
        self._outer_layout.addWidget(self._label)
        self._update_label()
        self._label.setLayout(self._layout)

        self._value: str = ""
        self._display = QLabel()
        self._layout.addWidget(self._display)

        self._select_button = QToolButton()
        self._select_button.setIcon(material_icon(icon_name="settings", size=(20, 20)))
        self._show_select_button: bool = show_select_button
        self._layout.addWidget(self._select_button)
        self._display.setMinimumHeight(self._select_button.sizeHint().height())
        self.show_select_button = self._show_select_button

        self._select_button.clicked.connect(self.show_choice_dialog)
        self.get_bec_shortcuts()

        self._connected: bool = False
        self.connect_device()

    def _create_dialog(self):
        return ChoiceDialog(
            config=self._config,
            parent=self,
            client=self.client,
            show_config=self.show_config_signals,
            show_normal=self.show_normal_signals,
            show_hinted=self.show_hinted_signals,
        )

    @SafeSlot()
    def _process_dialog(self, device: str, signal: str):
        self.disconnect_device()
        self.device = device
        self.signal = signal
        self._update_label()
        self.connect_device()

    def show_choice_dialog(self):
        dialog = self._create_dialog()
        dialog.accepted_output.connect(self._process_dialog)
        dialog.open()
        return dialog

    def connect_device(self):
        """Subscribe to the Redis topic for the device to display"""
        if not self._connected and self._device and self._device in self.dev:
            self._connected = True
            self._readback_endpoint = MessageEndpoints.device_readback(self._device)
            self.bec_dispatcher.connect_slot(self.on_device_readback, self._readback_endpoint)
            self._manual_read()
            self.set_display_value(self._value)

    def disconnect_device(self):
        """Unsubscribe from the Redis topic for the device to display"""
        if self._connected:
            self._connected = False
            self.bec_dispatcher.disconnect_slot(self.on_device_readback, self._readback_endpoint)

    def _manual_read(self):
        if self._device is None or not isinstance(
            (device := self.dev.get(self._device)), Device | Signal
        ):
            self._units = ""
            self._value = "__"
            return
        signal: Signal = (
            getattr(device, self.signal, None) if isinstance(device, Device) else device
        )
        if not isinstance(signal, Signal):  # Avoid getting other attributes of device, e.g. methods
            signal = None
        if signal is None:
            self._units = ""
            self._value = "__"
            return
        self._value = signal.get()
        self._units = signal.get_device_config().get("egu", "")

    @SafeSlot(dict, dict)
    def on_device_readback(self, msg: dict, metadata: dict) -> None:
        """
        Update the display with the new value.
        """
        try:
            signal_to_read = self._patch_hinted_signal()
            self._value = msg["signals"][signal_to_read]["value"]
            self.set_display_value(self._value)
        except Exception as e:
            self._display.setText("ERROR!")
            self._display.setToolTip(
                f"Error processing incoming reading: {msg}, handled with exception: {''.join(traceback.format_exception(e))}"
            )

    def _patch_hinted_signal(self):
        if self.dev[self._device]._info["signals"] == {}:
            return self._signal
        signal_info = self.dev[self._device]._info["signals"][self._signal]
        return (
            signal_info["obj_name"] if signal_info["kind_str"] == Kind.hinted.name else self._signal
        )

    @SafeProperty(str)
    def device(self) -> str:
        """The device from which to select a signal"""
        return self._device or "Not set!"

    @device.setter
    def device(self, value: str) -> None:
        self.disconnect_device()
        self._device = value
        self._config.device = value
        self.connect_device()
        self._update_label()

    @SafeProperty(str)
    def signal(self) -> str:
        """The signal to display"""
        return self._signal or "Not set!"

    @signal.setter
    def signal(self, value: str) -> None:
        self.disconnect_device()
        self._signal = value
        self._config.default = value
        self.connect_device()
        self._update_label()

    @SafeProperty(bool)
    def show_select_button(self) -> bool:
        """Show the button to select the signal to display"""
        return self._show_select_button

    @show_select_button.setter
    def show_select_button(self, value: bool) -> None:
        self._show_select_button = value
        self._select_button.setVisible(value)

    @SafeProperty(bool)
    def show_default_units(self) -> bool:
        """Show default units obtained from the signal alongside it"""
        return self._show_default_units

    @show_default_units.setter
    def show_default_units(self, value: bool) -> None:
        self._show_default_units = value
        self.set_display_value(self._value)

    @SafeProperty(str)
    def custom_label(self) -> str:
        """Use a cusom label rather than the signal name"""
        return self._custom_label

    @custom_label.setter
    def custom_label(self, value: str) -> None:
        self._custom_label = value
        self._update_label()

    @SafeProperty(str)
    def custom_units(self) -> str:
        """Use a custom unit string"""
        return self._custom_units

    @custom_units.setter
    def custom_units(self, value: str) -> None:
        self._custom_units = value
        self.set_display_value(self._value)

    @SafeProperty(int)
    def decimal_places(self) -> int:
        """Format to a given number of decimal_places. Set to 0 to disable."""
        return self._decimal_places

    @decimal_places.setter
    def decimal_places(self, value: int) -> None:
        self._decimal_places = value
        self._update_label()

    @SafeProperty(bool)
    def show_hinted_signals(self) -> bool:
        """In the signal selection menu, show hinted signals"""
        return self._show_hinted_signals

    @show_hinted_signals.setter
    def show_hinted_signals(self, value: bool) -> None:
        self._show_hinted_signals = value

    @SafeProperty(bool)
    def show_config_signals(self) -> bool:
        """In the signal selection menu, show config signals"""
        return self._show_config_signals

    @show_config_signals.setter
    def show_config_signals(self, value: bool) -> None:
        self._show_config_signals = value

    @SafeProperty(bool)
    def show_normal_signals(self) -> bool:
        """In the signal selection menu, show normal signals"""
        return self._show_normal_signals

    @show_normal_signals.setter
    def show_normal_signals(self, value: bool) -> None:
        self._show_normal_signals = value

    def _format_value(self, value: str):
        if self._decimal_places == 0:
            return value
        try:
            return f"{float(value):0.{self._decimal_places}f}"
        except ValueError:
            return value

    @SafeSlot(str)
    def set_display_value(self, value: str):
        """Set the display to a given value, appending the units if specified"""
        self._display.setText(f"{self._format_value(value)}{self._units_string}")
        self._display.setToolTip("")

    @property
    def _units_string(self):
        if self.custom_units or self._show_default_units:
            return f" {self.custom_units or self._default_units or ''}"
        return ""

    @property
    def _default_units(self) -> str:
        return self._units

    @property
    def _default_label(self) -> str:
        return (
            str(self._signal) if self._device == self._signal else f"{self._device} {self._signal}"
        )

    def _update_label(self):
        self._label.setTitle(
            self._custom_label if self._custom_label else f"{self._default_label}:"
        )


if __name__ == "__main__":

    app = QApplication(sys.argv)
    w = QWidget()
    w.setLayout(QVBoxLayout())
    w.layout().addWidget(
        SignalLabel(
            device="samx",
            signal="readback",
            custom_label="custom label:",
            custom_units=" m/s/s",
            show_select_button=False,
        )
    )
    w.layout().addWidget(SignalLabel(device="samy", signal="readback", show_default_units=True))
    l = SignalLabel()
    l.device = "bpm4i"
    l.signal = "bpm4i"
    w.layout().addWidget(l)
    w.show()
    sys.exit(app.exec_())



from __future__ import annotations

from typing import Literal, Optional

import pyqtgraph as pg
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from pydantic import Field, field_validator
from pydantic_core import PydanticCustomError
from qtpy import QtCore, QtGui
from qtpy.QtCore import QSize, Slot
from qtpy.QtWidgets import QSizePolicy, QWidget

from bec_widgets.utils import Colors, ConnectionConfig, EntryValidator
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.widgets.progress.ring_progress_bar.ring import Ring, RingConfig

logger = bec_logger.logger


class RingProgressBarConfig(ConnectionConfig):
    color_map: Optional[str] = Field(
        "plasma", description="Color scheme for the progress bars.", validate_default=True
    )
    min_number_of_bars: int | None = Field(
        1, description="Minimum number of progress bars to display."
    )
    max_number_of_bars: int | None = Field(
        10, description="Maximum number of progress bars to display."
    )
    num_bars: int | None = Field(1, description="Number of progress bars to display.")
    gap: int | None = Field(20, description="Gap between progress bars.")
    auto_updates: bool | None = Field(
        True, description="Enable or disable updates based on scan queue status."
    )
    rings: list[RingConfig] | None = Field([], description="List of ring configurations.")

    @field_validator("num_bars")
    @classmethod
    def validate_num_bars(cls, v, values):
        min_number_of_bars = values.data.get("min_number_of_bars", None)
        max_number_of_bars = values.data.get("max_number_of_bars", None)
        if min_number_of_bars is not None and max_number_of_bars is not None:
            logger.info(
                f"Number of bars adjusted to be between defined min:{min_number_of_bars} and max:{max_number_of_bars} number of bars."
            )
            v = max(min_number_of_bars, min(v, max_number_of_bars))
            return v

    @field_validator("rings")
    @classmethod
    def validate_rings(cls, v, values):
        if v is not None and v is not []:
            num_bars = values.data.get("num_bars", None)
            if len(v) != num_bars:
                raise PydanticCustomError(
                    "different number of configs",
                    f"Length of rings configuration ({len(v)}) does not match the number of bars ({num_bars}).",
                    {"wrong_value": len(v)},
                )
            indices = [ring.index for ring in v]
            if sorted(indices) != list(range(len(indices))):
                raise PydanticCustomError(
                    "wrong indices",
                    f"Indices of ring configurations must be unique and in order from 0 to num_bars {num_bars}.",
                    {"wrong_value": indices},
                )
        return v

    _validate_colormap = field_validator("color_map")(Colors.validate_color_map)


class RingProgressBar(BECWidget, QWidget):
    """
    Show the progress of devices, scans or custom values in the form of ring progress bars.
    """

    PLUGIN = True
    ICON_NAME = "track_changes"
    USER_ACCESS = [
        "_get_all_rpc",
        "_rpc_id",
        "_config_dict",
        "rings",
        "update_config",
        "add_ring",
        "remove_ring",
        "set_precision",
        "set_min_max_values",
        "set_number_of_bars",
        "set_value",
        "set_colors_from_map",
        "set_colors_directly",
        "set_line_widths",
        "set_gap",
        "set_diameter",
        "reset_diameter",
        "enable_auto_updates",
    ]

    def __init__(
        self,
        parent=None,
        config: RingProgressBarConfig | dict | None = None,
        client=None,
        gui_id: str | None = None,
        num_bars: int | None = None,
        **kwargs,
    ):
        if config is None:
            config = RingProgressBarConfig(widget_class=self.__class__.__name__)
            self.config = config
        else:
            if isinstance(config, dict):
                config = RingProgressBarConfig(**config, widget_class=self.__class__.__name__)
            self.config = config
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)

        self.get_bec_shortcuts()
        self.entry_validator = EntryValidator(self.dev)

        self.RID = None

        # For updating bar behaviour
        self._auto_updates = True
        self._rings = []

        if num_bars is not None:
            self.config.num_bars = max(
                self.config.min_number_of_bars, min(num_bars, self.config.max_number_of_bars)
            )
        self.initialize_bars()

        self.enable_auto_updates(self.config.auto_updates)

    @property
    def rings(self) -> list[Ring]:
        """Returns a list of all rings in the progress bar."""
        return self._rings

    @rings.setter
    def rings(self, value: list[Ring]):
        self._rings = value

    def update_config(self, config: RingProgressBarConfig | dict):
        """
        Update the configuration of the widget.

        Args:
            config(SpiralProgressBarConfig|dict): Configuration to update.
        """
        if isinstance(config, dict):
            config = RingProgressBarConfig(**config, widget_class=self.__class__.__name__)
        self.config = config
        self.clear_all()

    def initialize_bars(self):
        """
        Initialize the progress bars.
        """
        start_positions = [90 * 16] * self.config.num_bars
        directions = [-1] * self.config.num_bars

        self.config.rings = [
            RingConfig(
                widget_class="Ring",
                index=i,
                start_positions=start_positions[i],
                directions=directions[i],
            )
            for i in range(self.config.num_bars)
        ]
        self._rings = [Ring(parent=self, config=config) for config in self.config.rings]

        if self.config.color_map:
            self.set_colors_from_map(self.config.color_map)

        min_size = self._calculate_minimum_size()
        self.setMinimumSize(min_size)
        # Set outer ring to listen to scan progress
        self.rings[0].set_update(mode="scan")
        self.update()

    def add_ring(self, **kwargs) -> Ring:
        """
        Add a new progress bar.

        Args:
            **kwargs: Keyword arguments for the new progress bar.

        Returns:
            Ring: Ring object.
        """
        if self.config.num_bars < self.config.max_number_of_bars:
            ring_index = self.config.num_bars
            ring_config = RingConfig(
                widget_class="Ring",
                index=ring_index,
                start_positions=90 * 16,
                directions=-1,
                **kwargs,
            )
            ring = Ring(parent=self, config=ring_config)
            self.config.num_bars += 1
            self._rings.append(ring)
            self.config.rings.append(ring.config)
            if self.config.color_map:
                self.set_colors_from_map(self.config.color_map)
            base_line_width = self._rings[ring.config.index].config.line_width
            self.set_line_widths(base_line_width, ring.config.index)
            self.update()
            return ring

    def remove_ring(self, index: int):
        """
        Remove a progress bar by index.

        Args:
            index(int): Index of the progress bar to remove.
        """
        ring = self._find_ring_by_index(index)
        self._cleanup_ring(ring)
        self.update()

    def _cleanup_ring(self, ring: Ring) -> None:
        ring.reset_connection()
        self._rings.remove(ring)
        self.config.rings.remove(ring.config)
        self.config.num_bars -= 1
        self._reindex_rings()
        if self.config.color_map:
            self.set_colors_from_map(self.config.color_map)
        # Remove ring from rpc, afterwards call close event.
        ring.rpc_register.remove_rpc(ring)
        ring.deleteLater()
        # del ring

    def _reindex_rings(self):
        """
        Reindex the progress bars.
        """
        for i, ring in enumerate(self._rings):
            ring.config.index = i

    def set_precision(self, precision: int, bar_index: int = None):
        """
        Set the precision for the progress bars. If bar_index is not provide, the precision will be set for all progress bars.

        Args:
            precision(int): Precision for the progress bars.
            bar_index(int): Index of the progress bar to set the precision for. If provided, only a single precision can be set.
        """
        if bar_index is not None:
            bar_index = self._bar_index_check(bar_index)
            ring = self._find_ring_by_index(bar_index)
            ring.config.precision = precision
        else:
            for ring in self._rings:
                ring.config.precision = precision
        self.update()

    def set_min_max_values(
        self,
        min_values: int | float | list[int | float],
        max_values: int | float | list[int | float],
    ):
        """
        Set the minimum and maximum values for the progress bars.

        Args:
            min_values(int|float | list[float]): Minimum value(s) for the progress bars. If multiple progress bars are displayed, provide a list of minimum values for each progress bar.
            max_values(int|float | list[float]): Maximum value(s) for the progress bars. If multiple progress bars are displayed, provide a list of maximum values for each progress bar.
        """
        if isinstance(min_values, int) or isinstance(min_values, float):
            min_values = [min_values]
        if isinstance(max_values, int) or isinstance(max_values, float):
            max_values = [max_values]
        min_values = self._adjust_list_to_bars(min_values)
        max_values = self._adjust_list_to_bars(max_values)
        for ring, min_value, max_value in zip(self._rings, min_values, max_values):
            ring.set_min_max_values(min_value, max_value)
        self.update()

    def set_number_of_bars(self, num_bars: int):
        """
        Set the number of progress bars to display.

        Args:
            num_bars(int): Number of progress bars to display.
        """
        num_bars = max(
            self.config.min_number_of_bars, min(num_bars, self.config.max_number_of_bars)
        )
        current_num_bars = self.config.num_bars

        if num_bars > current_num_bars:
            for i in range(current_num_bars, num_bars):
                new_ring_config = RingConfig(
                    widget_class="Ring", index=i, start_positions=90 * 16, directions=-1
                )
                self.config.rings.append(new_ring_config)
                new_ring = Ring(parent=self, config=new_ring_config)
                self._rings.append(new_ring)

        elif num_bars < current_num_bars:
            for i in range(current_num_bars - 1, num_bars - 1, -1):
                self.remove_ring(i)

        self.config.num_bars = num_bars

        if self.config.color_map:
            self.set_colors_from_map(self.config.color_map)

        base_line_width = self._rings[0].config.line_width
        self.set_line_widths(base_line_width)

        self.update()

    def set_value(self, values: int | list, ring_index: int = None):
        """
        Set the values for the progress bars.

        Args:
            values(int | tuple): Value(s) for the progress bars. If multiple progress bars are displayed, provide a tuple of values for each progress bar.
            ring_index(int): Index of the progress bar to set the value for. If provided, only a single value can be set.

        Examples:
            >>> SpiralProgressBar.set_value(50)
            >>> SpiralProgressBar.set_value([30, 40, 50]) # (outer, middle, inner)
            >>> SpiralProgressBar.set_value(60, bar_index=1) # Set the value for the middle progress bar.
        """
        if ring_index is not None:
            ring = self._find_ring_by_index(ring_index)
            if isinstance(values, list):
                values = values[0]
                logger.warning(
                    f"Warning: Only a single value can be set for a single progress bar. Using the first value in the list {values}"
                )
            ring.set_value(values)
        else:
            if isinstance(values, int):
                values = [values]
                values = self._adjust_list_to_bars(values)
            for ring, value in zip(self._rings, values):
                ring.set_value(value)
        self.update()

    def set_colors_from_map(self, colormap, color_format: Literal["RGB", "HEX"] = "RGB"):
        """
        Set the colors for the progress bars from a colormap.

        Args:
            colormap(str): Name of the colormap.
            color_format(Literal["RGB","HEX"]): Format of the returned colors ('RGB', 'HEX').
        """
        if colormap not in pg.colormap.listMaps():
            raise ValueError(
                f"Colormap '{colormap}' not found in the current installation of pyqtgraph"
            )
        colors = Colors.golden_angle_color(colormap, self.config.num_bars, color_format)
        self.set_colors_directly(colors)
        self.config.color_map = colormap
        self.update()

    def set_colors_directly(self, colors: list[str | tuple] | str | tuple, bar_index: int = None):
        """
        Set the colors for the progress bars directly.

        Args:
            colors(list[str | tuple] | str | tuple): Color(s) for the progress bars. If multiple progress bars are displayed, provide a list of colors for each progress bar.
            bar_index(int): Index of the progress bar to set the color for. If provided, only a single color can be set.
        """
        if bar_index is not None and isinstance(colors, (str, tuple)):
            bar_index = self._bar_index_check(bar_index)
            ring = self._find_ring_by_index(bar_index)
            ring.set_color(colors)
        else:
            if isinstance(colors, (str, tuple)):
                colors = [colors]
            colors = self._adjust_list_to_bars(colors)
            for ring, color in zip(self._rings, colors):
                ring.set_color(color)
        self.update()

    def set_line_widths(self, widths: int | list[int], bar_index: int = None):
        """
        Set the line widths for the progress bars.

        Args:
            widths(int | list[int]): Line width(s) for the progress bars. If multiple progress bars are displayed, provide a list of line widths for each progress bar.
            bar_index(int): Index of the progress bar to set the line width for. If provided, only a single line width can be set.
        """
        if bar_index is not None:
            bar_index = self._bar_index_check(bar_index)
            ring = self._find_ring_by_index(bar_index)
            if isinstance(widths, list):
                widths = widths[0]
                logger.warning(
                    f"Warning: Only a single line width can be set for a single progress bar. Using the first value in the list {widths}"
                )
            ring.set_line_width(widths)
        else:
            if isinstance(widths, int):
                widths = [widths]
            widths = self._adjust_list_to_bars(widths)
            self.config.gap = max(widths) * 2
            for ring, width in zip(self._rings, widths):
                ring.set_line_width(width)
        min_size = self._calculate_minimum_size()
        self.setMinimumSize(min_size)
        self.update()

    def set_gap(self, gap: int):
        """
        Set the gap between the progress bars.

        Args:
            gap(int): Gap between the progress bars.
        """
        self.config.gap = gap
        self.update()

    def set_diameter(self, diameter: int):
        """
        Set the diameter of the widget.

        Args:
            diameter(int): Diameter of the widget.
        """
        size = QSize(diameter, diameter)
        self.resize(size)
        self.setFixedSize(size)

    def _find_ring_by_index(self, index: int) -> Ring:
        """
        Find the ring by index.

        Args:
            index(int): Index of the ring.

        Returns:
            Ring: Ring object.
        """
        found_ring = None
        for ring in self._rings:
            if ring.config.index == index:
                found_ring = ring
                break
        if found_ring is None:
            raise ValueError(f"Ring with index {index} not found.")
        return found_ring

    def enable_auto_updates(self, enable: bool = True):
        """
        Enable or disable updates based on scan status. Overrides manual updates.
        The behaviour of the whole progress bar widget will be driven by the scan queue status.

        Args:
            enable(bool): True or False.

        Returns:
            bool: True if scan segment updates are enabled.
        """

        self._auto_updates = enable
        if enable is True:
            self.bec_dispatcher.connect_slot(
                self.on_scan_queue_status, MessageEndpoints.scan_queue_status()
            )
        else:
            self.bec_dispatcher.disconnect_slot(
                self.on_scan_queue_status, MessageEndpoints.scan_queue_status()
            )
        return self._auto_updates

    @Slot(dict, dict)
    def on_scan_queue_status(self, msg, meta):
        """
        Slot to handle scan queue status messages. Decides what update to perform based on the scan queue status.

        Args:
            msg(dict): Message from the BEC.
            meta(dict): Metadata from the BEC.
        """
        primary_queue = msg.get("queue").get("primary")
        info = primary_queue.get("info", None)

        if info:
            active_request_block = info[0].get("active_request_block", None)
            if active_request_block:
                report_instructions = active_request_block.get("report_instructions", None)
                if report_instructions:
                    instruction_type = list(report_instructions[0].keys())[0]
                    if instruction_type == "scan_progress":
                        self._hook_scan_progress(ring_index=0)
                    elif instruction_type == "readback":
                        devices = report_instructions[0].get("readback").get("devices")
                        start = report_instructions[0].get("readback").get("start")
                        end = report_instructions[0].get("readback").get("end")
                        if self.config.num_bars != len(devices):
                            self.set_number_of_bars(len(devices))
                        for index, device in enumerate(devices):
                            self._hook_readback(index, device, start[index], end[index])
                    else:
                        logger.error(f"{instruction_type} not supported yet.")

                    # elif instruction_type == "device_progress":
                    #     print("hook device_progress")

    def _hook_scan_progress(self, ring_index: int = None):
        """
        Hook the scan progress to the progress bars.

        Args:
            ring_index(int): Index of the progress bar to hook the scan progress to.
        """
        if ring_index is not None:
            ring = self._find_ring_by_index(ring_index)
        else:
            ring = self._rings[0]

        if ring.config.connections.slot == "on_scan_progress":
            return
        else:
            ring.set_connections("on_scan_progress", MessageEndpoints.scan_progress())

    def _hook_readback(self, bar_index: int, device: str, min: float | int, max: float | int):
        """
        Hook the readback values to the progress bars.

        Args:
            bar_index(int): Index of the progress bar to hook the readback values to.
            device(str): Device to readback values from.
            min(float|int): Minimum value for the progress bar.
            max(float|int): Maximum value for the progress bar.
        """
        ring = self._find_ring_by_index(bar_index)
        ring.set_min_max_values(min, max)
        endpoint = MessageEndpoints.device_readback(device)
        ring.set_connections("on_device_readback", endpoint)

    def _adjust_list_to_bars(self, items: list) -> list:
        """
        Utility method to adjust the list of parameters to match the number of progress bars.

        Args:
            items(list): List of parameters for the progress bars.

        Returns:
            list: List of parameters for the progress bars.
        """
        if items is None:
            raise ValueError(
                "Items cannot be None. Please provide a list for parameters for the progress bars."
            )
        if not isinstance(items, list):
            items = [items]
        if len(items) < self.config.num_bars:
            last_item = items[-1]
            items.extend([last_item] * (self.config.num_bars - len(items)))
        elif len(items) > self.config.num_bars:
            items = items[: self.config.num_bars]
        return items

    def _bar_index_check(self, bar_index: int):
        """
        Utility method to check if the bar index is within the range of the number of progress bars.

        Args:
            bar_index(int): Index of the progress bar to set the value for.
        """
        if not (0 <= bar_index < self.config.num_bars):
            raise ValueError(
                f"bar_index {bar_index} out of range of number of bars {self.config.num_bars}."
            )
        return bar_index

    def paintEvent(self, event):
        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)
        size = min(self.width(), self.height())
        rect = QtCore.QRect(0, 0, size, size)
        rect.adjust(
            max(ring.config.line_width for ring in self._rings),
            max(ring.config.line_width for ring in self._rings),
            -max(ring.config.line_width for ring in self._rings),
            -max(ring.config.line_width for ring in self._rings),
        )

        for i, ring in enumerate(self._rings):
            # Background arc
            painter.setPen(
                QtGui.QPen(ring.background_color, ring.config.line_width, QtCore.Qt.SolidLine)
            )
            offset = self.config.gap * i
            adjusted_rect = QtCore.QRect(
                rect.left() + offset,
                rect.top() + offset,
                rect.width() - 2 * offset,
                rect.height() - 2 * offset,
            )
            painter.drawArc(adjusted_rect, ring.config.start_position, 360 * 16)

            # Foreground arc
            pen = QtGui.QPen(ring.color, ring.config.line_width, QtCore.Qt.SolidLine)
            pen.setCapStyle(QtCore.Qt.RoundCap)
            painter.setPen(pen)
            proportion = (ring.config.value - ring.config.min_value) / (
                (ring.config.max_value - ring.config.min_value) + 1e-3
            )
            angle = int(proportion * 360 * 16 * ring.config.direction)
            painter.drawArc(adjusted_rect, ring.start_position, angle)

    def reset_diameter(self):
        """
        Reset the fixed size of the widget.
        """
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        self.setMinimumSize(self._calculate_minimum_size())
        self.setMaximumSize(16777215, 16777215)

    def _calculate_minimum_size(self):
        """
        Calculate the minimum size of the widget.
        """
        if not self.config.rings:
            logger.warning("no rings to get size from setting size to 10x10")
            return QSize(10, 10)
        ring_widths = [self.config.rings[i].line_width for i in range(self.config.num_bars)]
        total_width = sum(ring_widths) + self.config.gap * (self.config.num_bars - 1)
        diameter = total_width * 2
        if diameter < 50:
            diameter = 50
        return QSize(diameter, diameter)

    def sizeHint(self):
        min_size = self._calculate_minimum_size()
        return min_size

    def clear_all(self):
        for ring in self._rings:
            ring.reset_connection()
        self._rings.clear()
        self.update()
        self.initialize_bars()

    def cleanup(self):
        self.bec_dispatcher.disconnect_slot(
            self.on_scan_queue_status, MessageEndpoints.scan_queue_status()
        )
        for ring in self._rings:
            self._cleanup_ring(ring)
        self._rings.clear()
        super().cleanup()



import os

import numpy as np
import pyqtgraph as pg
from bec_qthemes import material_icon
from qtpy.QtWidgets import (
    QApplication,
    QGroupBox,
    QHBoxLayout,
    QPushButton,
    QSplitter,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils import BECDispatcher
from bec_widgets.utils.widget_io import WidgetHierarchy as wh
from bec_widgets.widgets.containers.dock import BECDockArea
from bec_widgets.widgets.containers.layout_manager.layout_manager import LayoutManagerWidget
from bec_widgets.widgets.editors.jupyter_console.jupyter_console import BECJupyterConsole
from bec_widgets.widgets.plots.image.image import Image
from bec_widgets.widgets.plots.motor_map.motor_map import MotorMap
from bec_widgets.widgets.plots.multi_waveform.multi_waveform import MultiWaveform
from bec_widgets.widgets.plots.plot_base import PlotBase
from bec_widgets.widgets.plots.scatter_waveform.scatter_waveform import ScatterWaveform
from bec_widgets.widgets.plots.waveform.waveform import Waveform


class JupyterConsoleWindow(QWidget):  # pragma: no cover:
    """A widget that contains a Jupyter console linked to BEC Widgets with full API access (contains Qt and pyqtgraph API)."""

    def __init__(self, parent=None):
        super().__init__(parent)

        self._init_ui()

        # console push
        if self.console.inprocess is True:
            self.console.kernel_manager.kernel.shell.push(
                {
                    "np": np,
                    "pg": pg,
                    "wh": wh,
                    "dock": self.dock,
                    "im": self.im,
                    # "mi": self.mi,
                    # "mm": self.mm,
                    # "lm": self.lm,
                    # "btn1": self.btn1,
                    # "btn2": self.btn2,
                    # "btn3": self.btn3,
                    # "btn4": self.btn4,
                    # "btn5": self.btn5,
                    # "btn6": self.btn6,
                    # "pb": self.pb,
                    # "pi": self.pi,
                    # "wf": self.wf,
                    # "scatter": self.scatter,
                    # "scatter_mi": self.scatter,
                    # "mwf": self.mwf,
                }
            )

    def _init_ui(self):
        self.layout = QHBoxLayout(self)

        # Horizontal splitter
        splitter = QSplitter(self)
        self.layout.addWidget(splitter)

        tab_widget = QTabWidget(splitter)

        first_tab = QWidget()
        first_tab_layout = QVBoxLayout(first_tab)
        self.dock = BECDockArea(gui_id="dock")
        first_tab_layout.addWidget(self.dock)
        tab_widget.addTab(first_tab, "Dock Area")

        # third_tab = QWidget()
        # third_tab_layout = QVBoxLayout(third_tab)
        # self.lm = LayoutManagerWidget()
        # third_tab_layout.addWidget(self.lm)
        # tab_widget.addTab(third_tab, "Layout Manager Widget")
        #
        # fourth_tab = QWidget()
        # fourth_tab_layout = QVBoxLayout(fourth_tab)
        # self.pb = PlotBase()
        # self.pi = self.pb.plot_item
        # fourth_tab_layout.addWidget(self.pb)
        # tab_widget.addTab(fourth_tab, "PlotBase")
        #
        # tab_widget.setCurrentIndex(3)
        #
        group_box = QGroupBox("Jupyter Console", splitter)
        group_box_layout = QVBoxLayout(group_box)
        self.console = BECJupyterConsole(inprocess=True)
        group_box_layout.addWidget(self.console)
        #
        # # Some buttons for layout testing
        # self.btn1 = QPushButton("Button 1")
        # self.btn2 = QPushButton("Button 2")
        # self.btn3 = QPushButton("Button 3")
        # self.btn4 = QPushButton("Button 4")
        # self.btn5 = QPushButton("Button 5")
        # self.btn6 = QPushButton("Button 6")
        #
        # fifth_tab = QWidget()
        # fifth_tab_layout = QVBoxLayout(fifth_tab)
        # self.wf = Waveform()
        # fifth_tab_layout.addWidget(self.wf)
        # tab_widget.addTab(fifth_tab, "Waveform Next Gen")
        # tab_widget.setCurrentIndex(4)
        #
        sixth_tab = QWidget()
        sixth_tab_layout = QVBoxLayout(sixth_tab)
        self.im = Image(popups=True)
        self.mi = self.im.main_image
        sixth_tab_layout.addWidget(self.im)
        tab_widget.addTab(sixth_tab, "Image Next Gen")
        tab_widget.setCurrentIndex(1)
        #
        # seventh_tab = QWidget()
        # seventh_tab_layout = QVBoxLayout(seventh_tab)
        # self.scatter = ScatterWaveform()
        # self.scatter_mi = self.scatter.main_curve
        # self.scatter.plot("samx", "samy", "bpm4i")
        # seventh_tab_layout.addWidget(self.scatter)
        # tab_widget.addTab(seventh_tab, "Scatter Waveform")
        # tab_widget.setCurrentIndex(6)
        #
        # eighth_tab = QWidget()
        # eighth_tab_layout = QVBoxLayout(eighth_tab)
        # self.mm = MotorMap()
        # eighth_tab_layout.addWidget(self.mm)
        # tab_widget.addTab(eighth_tab, "Motor Map")
        # tab_widget.setCurrentIndex(7)
        #
        # ninth_tab = QWidget()
        # ninth_tab_layout = QVBoxLayout(ninth_tab)
        # self.mwf = MultiWaveform()
        # ninth_tab_layout.addWidget(self.mwf)
        # tab_widget.addTab(ninth_tab, "MultiWaveform")
        # tab_widget.setCurrentIndex(8)
        #
        # # add stuff to the new Waveform widget
        # self._init_waveform()
        #
        # self.setWindowTitle("Jupyter Console Window")

    def _init_waveform(self):
        self.wf.plot(y_name="bpm4i", y_entry="bpm4i", dap="GaussianModel")
        self.wf.plot(y_name="bpm3a", y_entry="bpm3a", dap="GaussianModel")

    def closeEvent(self, event):
        """Override to handle things when main window is closed."""
        self.dock.cleanup()
        self.dock.close()
        self.console.close()

        super().closeEvent(event)


if __name__ == "__main__":  # pragma: no cover
    import sys

    import bec_widgets

    module_path = os.path.dirname(bec_widgets.__file__)

    app = QApplication(sys.argv)
    app.setApplicationName("Jupyter Console")
    app.setApplicationDisplayName("Jupyter Console")
    icon = material_icon("terminal", color=(255, 255, 255, 255), filled=True)
    app.setWindowIcon(icon)

    bec_dispatcher = BECDispatcher(gui_id="jupyter_console")
    client = bec_dispatcher.client
    client.start()

    win = JupyterConsoleWindow()
    win.show()
    win.resize(1500, 800)

    app.aboutToQuit.connect(win.close)
    sys.exit(app.exec_())



"""
BECConsole is a Qt widget that runs a Bash shell.

BECConsole VT100 emulation is powered by Pyte,
(https://github.com/selectel/pyte).
"""

import collections
import fcntl
import html
import os
import pty
import re
import signal
import sys
import time

import pyte
from pygments.token import Token
from pyte.screens import History
from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import Property as pyqtProperty
from qtpy.QtCore import QSize, QSocketNotifier, Qt, QTimer
from qtpy.QtCore import Signal as pyqtSignal
from qtpy.QtGui import QClipboard, QColor, QPalette, QTextCursor
from qtpy.QtWidgets import QApplication, QHBoxLayout, QScrollBar, QSizePolicy

from bec_widgets.utils.error_popups import SafeSlot as Slot

ansi_colors = {
    "black": "#000000",
    "red": "#CD0000",
    "green": "#00CD00",
    "brown": "#996633",  # Brown, replacing the yellow
    "blue": "#0000EE",
    "magenta": "#CD00CD",
    "cyan": "#00CDCD",
    "white": "#E5E5E5",
    "brightblack": "#7F7F7F",
    "brightred": "#FF0000",
    "brightgreen": "#00FF00",
    "brightyellow": "#FFFF00",
    "brightblue": "#5C5CFF",
    "brightmagenta": "#FF00FF",
    "brightcyan": "#00FFFF",
    "brightwhite": "#FFFFFF",
}

control_keys_mapping = {
    QtCore.Qt.Key_A: b"\x01",  # Ctrl-A
    QtCore.Qt.Key_B: b"\x02",  # Ctrl-B
    QtCore.Qt.Key_C: b"\x03",  # Ctrl-C
    QtCore.Qt.Key_D: b"\x04",  # Ctrl-D
    QtCore.Qt.Key_E: b"\x05",  # Ctrl-E
    QtCore.Qt.Key_F: b"\x06",  # Ctrl-F
    QtCore.Qt.Key_G: b"\x07",  # Ctrl-G (Bell)
    QtCore.Qt.Key_H: b"\x08",  # Ctrl-H (Backspace)
    QtCore.Qt.Key_I: b"\x09",  # Ctrl-I (Tab)
    QtCore.Qt.Key_J: b"\x0a",  # Ctrl-J (Line Feed)
    QtCore.Qt.Key_K: b"\x0b",  # Ctrl-K (Vertical Tab)
    QtCore.Qt.Key_L: b"\x0c",  # Ctrl-L (Form Feed)
    QtCore.Qt.Key_M: b"\x0d",  # Ctrl-M (Carriage Return)
    QtCore.Qt.Key_N: b"\x0e",  # Ctrl-N
    QtCore.Qt.Key_O: b"\x0f",  # Ctrl-O
    QtCore.Qt.Key_P: b"\x10",  # Ctrl-P
    QtCore.Qt.Key_Q: b"\x11",  # Ctrl-Q
    QtCore.Qt.Key_R: b"\x12",  # Ctrl-R
    QtCore.Qt.Key_S: b"\x13",  # Ctrl-S
    QtCore.Qt.Key_T: b"\x14",  # Ctrl-T
    QtCore.Qt.Key_U: b"\x15",  # Ctrl-U
    QtCore.Qt.Key_V: b"\x16",  # Ctrl-V
    QtCore.Qt.Key_W: b"\x17",  # Ctrl-W
    QtCore.Qt.Key_X: b"\x18",  # Ctrl-X
    QtCore.Qt.Key_Y: b"\x19",  # Ctrl-Y
    QtCore.Qt.Key_Z: b"\x1a",  # Ctrl-Z
    QtCore.Qt.Key_Escape: b"\x1b",  # Ctrl-Escape
    QtCore.Qt.Key_Backslash: b"\x1c",  # Ctrl-\
    QtCore.Qt.Key_Underscore: b"\x1f",  # Ctrl-_
}

normal_keys_mapping = {
    QtCore.Qt.Key_Return: b"\n",
    QtCore.Qt.Key_Space: b" ",
    QtCore.Qt.Key_Enter: b"\n",
    QtCore.Qt.Key_Tab: b"\t",
    QtCore.Qt.Key_Backspace: b"\x08",
    QtCore.Qt.Key_Home: b"\x47",
    QtCore.Qt.Key_End: b"\x4f",
    QtCore.Qt.Key_Left: b"\x02",
    QtCore.Qt.Key_Up: b"\x10",
    QtCore.Qt.Key_Right: b"\x06",
    QtCore.Qt.Key_Down: b"\x0e",
    QtCore.Qt.Key_PageUp: b"\x49",
    QtCore.Qt.Key_PageDown: b"\x51",
    QtCore.Qt.Key_F1: b"\x1b\x31",
    QtCore.Qt.Key_F2: b"\x1b\x32",
    QtCore.Qt.Key_F3: b"\x1b\x33",
    QtCore.Qt.Key_F4: b"\x1b\x34",
    QtCore.Qt.Key_F5: b"\x1b\x35",
    QtCore.Qt.Key_F6: b"\x1b\x36",
    QtCore.Qt.Key_F7: b"\x1b\x37",
    QtCore.Qt.Key_F8: b"\x1b\x38",
    QtCore.Qt.Key_F9: b"\x1b\x39",
    QtCore.Qt.Key_F10: b"\x1b\x30",
    QtCore.Qt.Key_F11: b"\x45",
    QtCore.Qt.Key_F12: b"\x46",
}


def QtKeyToAscii(event):
    """
    Convert the Qt key event to the corresponding ASCII sequence for
    the terminal. This works fine for standard alphanumerical characters, but
    most other characters require terminal specific control sequences.

    The conversion below works for TERM="linux" terminals.
    """
    if sys.platform == "darwin":
        # special case for MacOS
        # /!\ Qt maps ControlModifier to CMD
        # CMD-C, CMD-V for copy/paste
        # CTRL-C and other modifiers -> key mapping
        if event.modifiers() == QtCore.Qt.MetaModifier:
            if event.key() == Qt.Key_Backspace:
                return control_keys_mapping.get(Qt.Key_W)
            return control_keys_mapping.get(event.key())
        elif event.modifiers() == QtCore.Qt.ControlModifier:
            if event.key() == Qt.Key_C:
                # copy
                return "copy"
            elif event.key() == Qt.Key_V:
                # paste
                return "paste"
            return None
        else:
            return normal_keys_mapping.get(event.key(), event.text().encode("utf8"))
    if event.modifiers() == QtCore.Qt.ControlModifier:
        return control_keys_mapping.get(event.key())
    else:
        return normal_keys_mapping.get(event.key(), event.text().encode("utf8"))


class Screen(pyte.HistoryScreen):
    def __init__(self, stdin_fd, cols, rows, historyLength):
        super().__init__(cols, rows, historyLength, ratio=1 / rows)
        self._fd = stdin_fd

    def write_process_input(self, data):
        """Response to CPR request (for example),
        this can be for other requests
        """
        try:
            os.write(self._fd, data.encode("utf-8"))
        except (IOError, OSError):
            pass

    def resize(self, lines, columns):
        lines = lines or self.lines
        columns = columns or self.columns

        if lines == self.lines and columns == self.columns:
            return  # No changes.

        self.dirty.clear()
        self.dirty.update(range(lines))

        self.save_cursor()
        if lines < self.lines:
            if lines <= self.cursor.y:
                nlines_to_move_up = self.lines - lines
                for i in range(nlines_to_move_up):
                    line = self.buffer[i]  # .pop(0)
                    self.history.top.append(line)
                self.cursor_position(0, 0)
                self.delete_lines(nlines_to_move_up)
                self.restore_cursor()
                self.cursor.y -= nlines_to_move_up
        else:
            self.restore_cursor()

        self.lines, self.columns = lines, columns
        self.history = History(
            self.history.top,
            self.history.bottom,
            1 / self.lines,
            self.history.size,
            self.history.position,
        )
        self.set_margins()


class Backend(QtCore.QObject):
    """
    Poll Bash.

    This class will run as a qsocketnotifier (started in ``_TerminalWidget``) and poll the
    file descriptor of the Bash terminal.
    """

    # Signals to communicate with ``_TerminalWidget``.
    dataReady = pyqtSignal(object)
    processExited = pyqtSignal()

    def __init__(self, fd, cols, rows):
        super().__init__()

        # File descriptor that connects to Bash process.
        self.fd = fd

        # Setup Pyte (hard coded display size for now).
        self.screen = Screen(self.fd, cols, rows, 10000)
        self.stream = pyte.ByteStream()
        self.stream.attach(self.screen)

        self.notifier = QSocketNotifier(fd, QSocketNotifier.Read)
        self.notifier.activated.connect(self._fd_readable)

    def _fd_readable(self):
        """
        Poll the Bash output, run it through Pyte, and notify
        """
        # Read the shell output until the file descriptor is closed.
        try:
            out = os.read(self.fd, 2**16)
        except OSError:
            self.processExited.emit()
            self.notifier.setEnabled(False)
            return

        # Feed output into Pyte's state machine and send the new screen
        # output to the GUI
        self.stream.feed(out)
        self.dataReady.emit(self.screen)


class BECConsole(QtWidgets.QWidget):
    """Container widget for the terminal text area"""

    PLUGIN = True
    ICON_NAME = "terminal"

    prompt = pyqtSignal(bool)

    def __init__(self, parent=None, cols=132):
        super().__init__(parent)

        self.term = _TerminalWidget(self, cols, rows=43)
        self.term.prompt.connect(self.prompt)  # forward signal from term to this widget

        self.scroll_bar = QScrollBar(Qt.Vertical, self)
        # self.scroll_bar.hide()
        layout = QHBoxLayout(self)
        layout.addWidget(self.term)
        layout.addWidget(self.scroll_bar)
        layout.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Expanding)

        pal = QPalette()
        self.set_bgcolor(pal.window().color())
        self.set_fgcolor(pal.windowText().color())
        self.term.set_scroll_bar(self.scroll_bar)
        self.set_cmd("bec --nogui")

        self._check_designer_timer = QTimer()
        self._check_designer_timer.timeout.connect(self.check_designer)
        self._check_designer_timer.start(1000)

    def minimumSizeHint(self):
        size = self.term.sizeHint()
        size.setWidth(size.width() + self.scroll_bar.width())
        return size

    def sizeHint(self):
        return self.minimumSizeHint()

    def check_designer(self, calls={"n": 0}):
        calls["n"] += 1
        if self.term.fd is not None:
            # already started
            self._check_designer_timer.stop()
        elif self.window().windowTitle().endswith("[Preview]"):
            # assuming Designer preview -> start
            self._check_designer_timer.stop()
            self.term.start()
        elif calls["n"] >= 3:
            # assuming not in Designer -> stop checking
            self._check_designer_timer.stop()

    def get_rows(self):
        return self.term.rows

    def set_rows(self, rows):
        self.term.rows = rows
        self.adjustSize()
        self.updateGeometry()

    def get_cols(self):
        return self.term.cols

    def set_cols(self, cols):
        self.term.cols = cols
        self.adjustSize()
        self.updateGeometry()

    def get_bgcolor(self):
        return QColor.fromString(self.term.bg_color)

    def set_bgcolor(self, color):
        self.term.bg_color = color.name(QColor.HexRgb)

    def get_fgcolor(self):
        return QColor.fromString(self.term.fg_color)

    def set_fgcolor(self, color):
        self.term.fg_color = color.name(QColor.HexRgb)

    def get_cmd(self):
        return self.term._cmd

    def set_cmd(self, cmd):
        self.term._cmd = cmd
        if self.term.fd is None:
            # not started yet
            self.term.clear()
            self.term.appendHtml(f"<h2>BEC Console - {repr(cmd)}</h2>")

    def start(self, deactivate_ctrl_d=True):
        self.term.start(deactivate_ctrl_d=deactivate_ctrl_d)

    def push(self, text, hit_return=False):
        """Push some text to the terminal"""
        return self.term.push(text, hit_return=hit_return)

    def execute_command(self, command):
        self.push(command, hit_return=True)

    def set_prompt_tokens(self, *tokens):
        """Prepare regexp to identify prompt, based on tokens

        Tokens are returned from get_ipython().prompts.in_prompt_tokens()
        """
        regex_parts = []
        for token_type, token_value in tokens:
            if token_type == Token.PromptNum:  # Handle dynamic prompt number
                regex_parts.append(r"[\d\?]+")  # Match one or more digits or '?'
            else:
                # Escape other prompt parts (e.g., "In [", "]: ")
                if not token_value:
                    regex_parts.append(".+?")  # arbitrary string
                else:
                    regex_parts.append(re.escape(token_value))

        # Combine into a single regex
        prompt_pattern = "".join(regex_parts)
        self.term._prompt_re = re.compile(prompt_pattern + r"\s*$")

    def terminate(self, timeout=10):
        self.term.stop(timeout=timeout)

    def send_ctrl_c(self, timeout=None):
        self.term.send_ctrl_c(timeout)

    cols = pyqtProperty(int, get_cols, set_cols)
    rows = pyqtProperty(int, get_rows, set_rows)
    bgcolor = pyqtProperty(QColor, get_bgcolor, set_bgcolor)
    fgcolor = pyqtProperty(QColor, get_fgcolor, set_fgcolor)
    cmd = pyqtProperty(str, get_cmd, set_cmd)


class _TerminalWidget(QtWidgets.QPlainTextEdit):
    """
    Start ``Backend`` process and render Pyte output as text.
    """

    prompt = pyqtSignal(bool)

    def __init__(self, parent, cols=125, rows=50, **kwargs):
        # regexp to match prompt
        self._prompt_re = None
        # last prompt
        self._prompt_str = None
        # process pid
        self.pid = None
        # file descriptor to communicate with the subprocess
        self.fd = None
        self.backend = None
        # command to execute
        self._cmd = ""
        # should ctrl-d be deactivated ? (prevent Python exit)
        self._deactivate_ctrl_d = False

        # Default colors
        pal = QPalette()
        self._fg_color = pal.text().color().name()
        self._bg_color = pal.base().color().name()

        # Specify the terminal size in terms of lines and columns.
        self._rows = rows
        self._cols = cols
        self.output = collections.deque()

        super().__init__(parent)

        self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Expanding)

        # Disable default scrollbars (we use our own, to be set via .set_scroll_bar())
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scroll_bar = None

        # Use Monospace fonts and disable line wrapping.
        self.setFont(QtGui.QFont("Courier", 9))
        self.setFont(QtGui.QFont("Monospace"))
        self.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        fmt = QtGui.QFontMetrics(self.font())
        char_width = fmt.width("w")
        self.setCursorWidth(char_width)

        self.adjustSize()
        self.updateGeometry()
        self.update_stylesheet()

    @property
    def bg_color(self):
        return self._bg_color

    @bg_color.setter
    def bg_color(self, hexcolor):
        self._bg_color = hexcolor
        self.update_stylesheet()

    @property
    def fg_color(self):
        return self._fg_color

    @fg_color.setter
    def fg_color(self, hexcolor):
        self._fg_color = hexcolor
        self.update_stylesheet()

    def update_stylesheet(self):
        self.setStyleSheet(
            f"QPlainTextEdit {{ border: 0; color: {self._fg_color}; background-color: {self._bg_color}; }} "
        )

    @property
    def rows(self):
        return self._rows

    @rows.setter
    def rows(self, rows: int):
        if self.backend is None:
            # not initialized yet, ok to change
            self._rows = rows
            self.adjustSize()
            self.updateGeometry()
        else:
            raise RuntimeError("Cannot change rows after console is started.")

    @property
    def cols(self):
        return self._cols

    @cols.setter
    def cols(self, cols: int):
        if self.fd is None:
            # not initialized yet, ok to change
            self._cols = cols
            self.adjustSize()
            self.updateGeometry()
        else:
            raise RuntimeError("Cannot change cols after console is started.")

    def start(self, deactivate_ctrl_d: bool = False):
        self._deactivate_ctrl_d = deactivate_ctrl_d

        self.update_term_size()

        # Start the Bash process
        self.pid, self.fd = self.fork_shell()

        if self.fd:
            # Create the ``Backend`` object
            self.backend = Backend(self.fd, self.cols, self.rows)
            self.backend.dataReady.connect(self.data_ready)
            self.backend.processExited.connect(self.process_exited)
        else:
            self.process_exited()

    def process_exited(self):
        self.fd = None
        self.clear()
        self.appendHtml(f"<br><h2>{repr(self._cmd)} - Process exited.</h2>")
        self.setReadOnly(True)

    def send_ctrl_c(self, wait_prompt=True, timeout=None):
        """Send CTRL-C to the process

        If wait_prompt=True (default), wait for a new prompt after CTRL-C
        If no prompt is displayed after 'timeout' seconds, TimeoutError is raised
        """
        os.kill(self.pid, signal.SIGINT)
        if wait_prompt:
            timeout_error = False
            if timeout:

                def set_timeout_error():
                    nonlocal timeout_error
                    timeout_error = True

                timeout_timer = QTimer()
                timeout_timer.singleShot(timeout * 1000, set_timeout_error)
            while self._prompt_str is None:
                QApplication.instance().process_events()
                if timeout_error:
                    raise TimeoutError(
                        f"CTRL-C: could not get back to prompt after {timeout} seconds."
                    )

    def _is_running(self):
        if os.waitpid(self.pid, os.WNOHANG) == (0, 0):
            return True
        return False

    def stop(self, kill=True, timeout=None):
        """Stop the running process

        SIGTERM is the default signal for terminating processes.

        If kill=True (default), SIGKILL will be sent if the process does not exit after timeout
        """
        # try to exit gracefully
        os.kill(self.pid, signal.SIGTERM)

        # wait until process is truly dead
        t0 = time.perf_counter()
        while self._is_running():
            time.sleep(1)
            if timeout is not None and time.perf_counter() - t0 > timeout:
                # still alive after 'timeout' seconds
                if kill:
                    # send SIGKILL and make a last check in loop
                    os.kill(self.pid, signal.SIGKILL)
                    kill = False
                else:
                    # still running after timeout...
                    raise TimeoutError(
                        f"Could not terminate process with pid: {self.pid} within timeout"
                    )
        self.process_exited()

    def data_ready(self, screen):
        """Handle new screen: redraw, set scroll bar max and slider, move cursor to its position

        This method is triggered via a signal from ``Backend``.
        """
        self.redraw_screen()
        self.adjust_scroll_bar()
        self.move_cursor()

    def minimumSizeHint(self):
        """Return minimum size for current cols and rows"""
        fmt = QtGui.QFontMetrics(self.font())
        char_width = fmt.width("w")
        char_height = fmt.height()
        width = char_width * self.cols
        height = char_height * self.rows
        return QSize(width, height)

    def sizeHint(self):
        return self.minimumSizeHint()

    def set_scroll_bar(self, scroll_bar):
        self.scroll_bar = scroll_bar
        self.scroll_bar.setMinimum(0)
        self.scroll_bar.valueChanged.connect(self.scroll_value_change)

    def scroll_value_change(self, value, old={"value": -1}):
        if self.backend is None:
            return
        if old["value"] == -1:
            old["value"] = self.scroll_bar.maximum()
        if value <= old["value"]:
            # scroll up
            # value is number of lines from the start
            nlines = old["value"] - value
            # history ratio gives prev_page == 1 line
            for i in range(nlines):
                self.backend.screen.prev_page()
        else:
            # scroll down
            nlines = value - old["value"]
            for i in range(nlines):
                self.backend.screen.next_page()
        old["value"] = value
        self.redraw_screen()

    def adjust_scroll_bar(self):
        sb = self.scroll_bar
        sb.valueChanged.disconnect(self.scroll_value_change)
        tmp = len(self.backend.screen.history.top) + len(self.backend.screen.history.bottom)
        sb.setMaximum(tmp if tmp > 0 else 0)
        sb.setSliderPosition(tmp if tmp > 0 else 0)
        # if tmp > 0:
        #    # show scrollbar, but delayed - prevent recursion with widget size change
        #    QTimer.singleShot(0, scrollbar.show)
        # else:
        #    QTimer.singleShot(0, scrollbar.hide)
        sb.valueChanged.connect(self.scroll_value_change)

    def write(self, data):
        try:
            os.write(self.fd, data)
        except (IOError, OSError):
            self.process_exited()

    @Slot(object)
    def keyPressEvent(self, event):
        """
        Redirect all keystrokes to the terminal process.
        """
        if self.fd is None:
            # not started
            return
        # Convert the Qt key to the correct ASCII code.
        if (
            self._deactivate_ctrl_d
            and event.modifiers() == QtCore.Qt.ControlModifier
            and event.key() == QtCore.Qt.Key_D
        ):
            return None

        code = QtKeyToAscii(event)
        if code == "copy":
            # MacOS only: CMD-C handling
            self.copy()
        elif code == "paste":
            # MacOS only: CMD-V handling
            self._push_clipboard()
        elif code is not None:
            self.write(code)

    def push(self, text, hit_return=False):
        """
        Write 'text' to terminal
        """
        self.write(text.encode("utf-8"))
        if hit_return:
            self.write(b"\n")

    def contextMenuEvent(self, event):
        if self.fd is None:
            return
        menu = self.createStandardContextMenu()
        for action in menu.actions():
            # remove all actions except copy and paste
            if "opy" in action.text():
                # redefine text without shortcut
                # since it probably clashes with control codes (like CTRL-C etc)
                action.setText("Copy")
                continue
            if "aste" in action.text():
                # redefine text without shortcut
                action.setText("Paste")
                # paste -> have to insert with self.push
                action.triggered.connect(self._push_clipboard)
                continue
            menu.removeAction(action)
        menu.exec_(event.globalPos())

    def _push_clipboard(self):
        clipboard = QApplication.instance().clipboard()
        self.push(clipboard.text())

    def move_cursor(self):
        textCursor = self.textCursor()
        textCursor.setPosition(0)
        textCursor.movePosition(
            QTextCursor.Down, QTextCursor.MoveAnchor, self.backend.screen.cursor.y
        )
        textCursor.movePosition(
            QTextCursor.Right, QTextCursor.MoveAnchor, self.backend.screen.cursor.x
        )
        self.setTextCursor(textCursor)

    def mouseReleaseEvent(self, event):
        if self.fd is None:
            return
        if event.button() == Qt.MiddleButton:
            # push primary selection buffer ("mouse clipboard") to terminal
            clipboard = QApplication.instance().clipboard()
            if clipboard.supportsSelection():
                self.push(clipboard.text(QClipboard.Selection))
            return None
        elif event.button() == Qt.LeftButton:
            # left button click
            textCursor = self.textCursor()
            if textCursor.selectedText():
                # mouse was used to select text -> nothing to do
                pass
            else:
                # a simple 'click', move scrollbar to end
                self.scroll_bar.setSliderPosition(self.scroll_bar.maximum())
                self.move_cursor()
                return None
        return super().mouseReleaseEvent(event)

    def redraw_screen(self):
        """
        Render the screen as formatted text into the widget.
        """
        screen = self.backend.screen

        # Clear the widget
        if screen.dirty:
            self.clear()
            while len(self.output) < (max(screen.dirty) + 1):
                self.output.append("")
            while len(self.output) > (max(screen.dirty) + 1):
                self.output.pop()

            # Prepare the HTML output
            for line_no in screen.dirty:
                line = text = ""
                style = old_style = ""
                old_idx = 0
                for idx, ch in screen.buffer[line_no].items():
                    text += " " * (idx - old_idx - 1)
                    old_idx = idx
                    style = f"{'background-color:%s;' % ansi_colors.get(ch.bg, ansi_colors['black']) if ch.bg!='default' else ''}{'color:%s;' % ansi_colors.get(ch.fg, ansi_colors['white']) if ch.fg!='default' else ''}{'font-weight:bold;' if ch.bold else ''}{'font-style:italic;' if ch.italics else ''}"
                    if style != old_style:
                        if old_style:
                            line += f"<span style={repr(old_style)}>{html.escape(text, quote=True)}</span>"
                        else:
                            line += html.escape(text, quote=True)
                        text = ""
                        old_style = style
                    text += ch.data
                if style:
                    line += f"<span style={repr(style)}>{html.escape(text, quote=True)}</span>"
                else:
                    line += html.escape(text, quote=True)
                # do a check at the cursor position:
                # it is possible x pos > output line length,
                # for example if last escape codes are "cursor forward" past end of text,
                # like IPython does for "..." prompt (in a block, like "for" loop or "while" for example)
                # In this case, cursor is at 12 but last text output is at 8 -> insert spaces
                if line_no == screen.cursor.y:
                    llen = len(screen.buffer[line_no])
                    if llen < screen.cursor.x:
                        line += " " * (screen.cursor.x - llen)
                self.output[line_no] = line
            # fill the text area with HTML contents in one go
            self.appendHtml(f"<pre>{chr(10).join(self.output)}</pre>")

            if self._prompt_re is not None:
                text_buf = self.toPlainText()
                prompt = self._prompt_re.search(text_buf)
                if prompt is None:
                    if self._prompt_str:
                        self.prompt.emit(False)
                    self._prompt_str = None
                else:
                    prompt_str = prompt.string.rstrip()
                    if prompt_str != self._prompt_str:
                        self._prompt_str = prompt_str
                        self.prompt.emit(True)

            # did updates, all clean
            screen.dirty.clear()

    def update_term_size(self):
        fmt = QtGui.QFontMetrics(self.font())
        char_width = fmt.width("w")
        char_height = fmt.height()
        self._cols = int(self.width() / char_width)
        self._rows = int(self.height() / char_height)

    def resizeEvent(self, event):
        self.update_term_size()
        if self.fd:
            self.backend.screen.resize(self._rows, self._cols)
            self.redraw_screen()
            self.adjust_scroll_bar()
            self.move_cursor()

    def wheelEvent(self, event):
        if not self.fd:
            return
        y = event.angleDelta().y()
        if y > 0:
            self.backend.screen.prev_page()
        else:
            self.backend.screen.next_page()
        self.redraw_screen()

    def fork_shell(self):
        """
        Fork the current process and execute bec in shell.
        """
        try:
            pid, fd = pty.fork()
        except (IOError, OSError):
            return False
        if pid == 0:
            try:
                ls = os.environ["LANG"].split(".")
            except KeyError:
                ls = []
            if len(ls) < 2:
                ls = ["en_US", "UTF-8"]
            os.putenv("COLUMNS", str(self.cols))
            os.putenv("LINES", str(self.rows))
            os.putenv("TERM", "linux")
            os.putenv("LANG", ls[0] + ".UTF-8")
            if not self._cmd:
                self._cmd = os.environ["SHELL"]
            cmd = self._cmd
            if isinstance(cmd, str):
                cmd = cmd.split()
            try:
                os.execvp(cmd[0], cmd)
            except (IOError, OSError):
                pass
            os._exit(0)
        else:
            # We are in the parent process.
            # Set file control
            fcntl.fcntl(fd, fcntl.F_SETFL, os.O_NONBLOCK)
            return pid, fd


if __name__ == "__main__":
    import os
    import sys

    from qtpy import QtGui, QtWidgets

    # Create the Qt application and console.
    app = QtWidgets.QApplication([])
    mainwin = QtWidgets.QMainWindow()
    title = "BECConsole"
    mainwin.setWindowTitle(title)

    console = BECConsole(mainwin)
    mainwin.setCentralWidget(console)

    def check_prompt(at_prompt):
        if at_prompt:
            print("NEW PROMPT")
        else:
            print("EXECUTING SOMETHING...")

    console.set_prompt_tokens(
        (Token.OutPromptNum, "•"),
        (Token.Prompt, ""),  # will match arbitrary string,
        (Token.Prompt, " ["),
        (Token.PromptNum, "3"),
        (Token.Prompt, "/"),
        (Token.PromptNum, "1"),
        (Token.Prompt, "] "),
        (Token.Prompt, "❯❯"),
    )
    console.prompt.connect(check_prompt)
    console.start()

    # Show widget and launch Qt's event loop.
    mainwin.show()
    sys.exit(app.exec_())



from __future__ import annotations

from functools import wraps
from threading import RLock
from typing import TYPE_CHECKING, Callable
from weakref import WeakValueDictionary

from bec_lib.logger import bec_logger
from qtpy.QtCore import QObject

if TYPE_CHECKING:  # pragma: no cover
    from bec_widgets.utils.bec_connector import BECConnector
    from bec_widgets.utils.bec_widget import BECWidget
    from bec_widgets.widgets.containers.dock.dock import BECDock
    from bec_widgets.widgets.containers.dock.dock_area import BECDockArea

logger = bec_logger.logger


def broadcast_update(func):
    """
    Decorator to broadcast updates to the RPCRegister whenever a new RPC object is added or removed.
    If class attribute _skip_broadcast is set to True, the broadcast will be skipped
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        result = func(self, *args, **kwargs)
        self.broadcast()
        return result

    return wrapper


class RPCRegister:
    """
    A singleton class that keeps track of all the RPC objects registered in the system for CLI usage.
    """

    _instance = None
    _initialized = False

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(RPCRegister, cls).__new__(cls)
            cls._initialized = False
        return cls._instance

    def __init__(self):
        if self._initialized:
            return
        self._rpc_register = WeakValueDictionary()
        self._broadcast_on_hold = RPCRegisterBroadcast(self)
        self._lock = RLock()
        self._skip_broadcast = False
        self._initialized = True
        self.callbacks = []

    @classmethod
    def delayed_broadcast(cls):
        """
        Delay the broadcast of the update to all the callbacks.
        """
        register = cls()
        return register._broadcast_on_hold

    @broadcast_update
    def add_rpc(self, rpc: BECConnector):
        """
        Add an RPC object to the register.

        Args:
            rpc(QObject): The RPC object to be added to the register.
        """
        if not hasattr(rpc, "gui_id"):
            raise ValueError("RPC object must have a 'gui_id' attribute.")
        self._rpc_register[rpc.gui_id] = rpc

    @broadcast_update
    def remove_rpc(self, rpc: BECConnector):
        """
        Remove an RPC object from the register.

        Args:
            rpc(str): The RPC object to be removed from the register.
        """
        if not hasattr(rpc, "gui_id"):
            raise ValueError(f"RPC object {rpc} must have a 'gui_id' attribute.")
        self._rpc_register.pop(rpc.gui_id, None)

    def get_rpc_by_id(self, gui_id: str) -> QObject | None:
        """
        Get an RPC object by its ID.

        Args:
            gui_id(str): The ID of the RPC object to be retrieved.

        Returns:
            QObject | None: The RPC object with the given ID or None
        """
        rpc_object = self._rpc_register.get(gui_id, None)
        return rpc_object

    def list_all_connections(self) -> dict:
        """
        List all the registered RPC objects.

        Returns:
            dict: A dictionary containing all the registered RPC objects.
        """
        with self._lock:
            connections = dict(self._rpc_register)
        return connections

    def get_names_of_rpc_by_class_type(
        self, cls: type[BECWidget] | type[BECConnector] | type[BECDock] | type[BECDockArea]
    ) -> list[str]:
        """Get all the names of the widgets.

        Args:
            cls(BECWidget | BECConnector): The class of the RPC object to be retrieved.
        """
        # This retrieves any rpc objects that are subclass of BECWidget,
        # i.e. curve and image items are excluded
        widgets = [rpc for rpc in self._rpc_register.values() if isinstance(rpc, cls)]
        return [widget.object_name for widget in widgets]

    def broadcast(self):
        """
        Broadcast the update to all the callbacks.
        """

        if self._skip_broadcast:
            return
        connections = self.list_all_connections()
        for callback in self.callbacks:
            callback(connections)

    def object_is_registered(self, obj: BECConnector) -> bool:
        """
        Check if an object is registered in the RPC register.

        Args:
            obj(QObject): The object to check.

        Returns:
            bool: True if the object is registered, False otherwise.
        """
        return obj.gui_id in self._rpc_register

    def add_callback(self, callback: Callable[[dict], None]):
        """
        Add a callback that will be called whenever the registry is updated.

        Args:
            callback(Callable[[dict], None]): The callback to be added. It should accept a dictionary of all the
            registered RPC objects as an argument.
        """
        self.callbacks.append(callback)

    @classmethod
    def reset_singleton(cls):
        """
        Reset the singleton instance.
        """
        cls._instance = None
        cls._initialized = False


class RPCRegisterBroadcast:
    """Context manager for RPCRegister broadcast."""

    def __init__(self, rpc_register: RPCRegister) -> None:
        self.rpc_register = rpc_register
        self._call_depth = 0

    def __enter__(self):
        """Enter the context manager"""
        self._call_depth += 1  # Needed for nested calls
        self.rpc_register._skip_broadcast = True
        return self.rpc_register

    def __exit__(self, *exc):
        """Exit the context manager"""

        self._call_depth -= 1  # Remove nested calls
        if self._call_depth == 0:  # The Last one to exit is responsible for broadcasting
            self.rpc_register._skip_broadcast = False
        self.rpc_register.broadcast()



def rpc_public(func):
    func.rpc_public = True  # Mark the function for later processing by the class decorator
    return func


def register_rpc_methods(cls):
    """
    Class decorator to scan for rpc_public methods and add them to USER_ACCESS.
    """
    if not hasattr(cls, "USER_ACCESS"):
        cls.USER_ACCESS = set()
    for name, method in cls.__dict__.items():
        if getattr(method, "rpc_public", False):
            cls.USER_ACCESS.add(name)
    return cls



"""Module for a PositionerGroup widget to control a positioner device."""

from __future__ import annotations

from bec_lib.device import Positioner
from bec_lib.logger import bec_logger
from qtpy.QtCore import QSize, Signal
from qtpy.QtWidgets import QGridLayout, QGroupBox, QVBoxLayout, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot
from bec_widgets.widgets.control.device_control.positioner_box import PositionerBox

logger = bec_logger.logger


class PositionerGroupBox(QGroupBox):
    PLUGIN = True
    position_update = Signal(float)

    def __init__(self, parent, dev_name):
        super().__init__(parent)

        self.device_name = dev_name

        QVBoxLayout(self)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.widget = PositionerBox(self, dev_name)
        self.widget.compact_view = True
        self.widget.expand_popup = False
        self.layout().addWidget(self.widget)
        self.widget.position_update.connect(self._on_position_update)
        self.widget.expand.connect(self._on_expand)
        self.setTitle(self.device_name)
        self.widget.force_update_readback()

    def _on_expand(self, expand):
        if expand:
            self.setTitle("")
            self.setFlat(True)
        else:
            self.setTitle(self.device_name)
            self.setFlat(False)

    def _on_position_update(self, pos: float):
        self.position_update.emit(pos)
        self.widget.label = f"%.{self.widget.dev[self.widget.device].precision}f" % pos

    def close(self):
        self.widget.close()
        super().close()


class PositionerGroup(BECWidget, QWidget):
    """Simple Widget to control a positioner in box form"""

    ICON_NAME = "grid_view"
    USER_ACCESS = ["set_positioners"]

    # Signal emitted to inform listeners about a position update of the first positioner
    position_update = Signal(float)
    # Signal emitted to inform listeners about (positioner, pos) updates
    device_position_update = Signal(str, float)

    def __init__(self, parent=None, **kwargs):
        """Initialize the widget.

        Args:
            parent: The parent widget.
        """
        super().__init__(parent=parent, **kwargs)

        self.get_bec_shortcuts()

        QGridLayout(self)
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._device_widgets = {}
        self._grid_ncols = 2

    def minimumSizeHint(self):
        return QSize(300, 30)

    @SafeSlot(str)
    def set_positioners(self, device_names: str):
        """Redraw grid with positioners from device_names string

        Device names must be separated by space
        """
        devs = device_names.split()
        for dev_name in devs:
            if not self._check_device_is_valid(dev_name):
                raise ValueError(f"{dev_name} is not a valid Positioner")
        for i, existing_widget in enumerate(self._device_widgets.values()):
            self.layout().removeWidget(existing_widget)
            existing_widget.position_update.disconnect(self._on_position_update)
            if i == 0:
                existing_widget.position_update.disconnect(self.position_update)
        for i, dev_name in enumerate(devs):
            widget = self._device_widgets.get(dev_name)
            if widget is None:
                widget = PositionerGroupBox(self, dev_name)
                self._device_widgets[dev_name] = widget
                widget.position_update.connect(self._on_position_update)
                if i == 0:
                    # only emit 'position_update' for the first positioner in grid
                    widget.position_update.connect(self.position_update)
            self.layout().addWidget(widget, i // self._grid_ncols, i % self._grid_ncols)
        to_remove = set(self._device_widgets) - set(devs)
        for dev_name in to_remove:
            self._device_widgets[dev_name].close()
            del self._device_widgets[dev_name]

    def _check_device_is_valid(self, device: str):
        """Check if the device is a positioner

        Args:
            device (str): The device name
        """
        if device not in self.dev:
            logger.info(f"Device {device} not found in the device list")
            return False
        if not isinstance(self.dev[device], Positioner):
            logger.info(f"Device {device} is not a positioner")
            return False
        return True

    def _on_position_update(self, pos: float):
        widget = self.sender()
        self.device_position_update.emit(widget.title(), pos)

    @SafeProperty(str)
    def devices_list(self):
        """Device names string separated by space"""
        return " ".join(self._device_widgets)

    @devices_list.setter
    def devices_list(self, device_names: str):
        """Set devices list from device names string separated by space"""
        devs = device_names.split()
        for dev_name in devs:
            if not self._check_device_is_valid(dev_name):
                return
        self.set_positioners(device_names)

    @SafeProperty(int)
    def grid_max_cols(self):
        """Max number of columns for widgets grid"""
        return self._grid_ncols

    @grid_max_cols.setter
    def grid_max_cols(self, ncols: int):
        """Set max number of columns for widgets grid"""
        self._grid_ncols = ncols
        self.set_positioners(self.devices_list)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication  # pylint: disable=ungrouped-imports

    app = QApplication(sys.argv)
    widget = PositionerGroup()
    widget.grid_max_cols = 3
    widget.set_positioners("samx samy samz")

    widget.show()
    sys.exit(app.exec_())



import os

from bec_lib.logger import bec_logger
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QPushButton, QTreeWidgetItem, QVBoxLayout, QWidget

from bec_widgets.utils import UILoader
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import get_accent_colors
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot

logger = bec_logger.logger


class LMFitDialog(BECWidget, QWidget):
    """Dialog for displaying the fit summary and params for LMFit DAP processes"""

    PLUGIN = True
    ICON_NAME = "monitoring"
    RPC = False
    # Signal to emit the currently selected fit curve_id
    selected_fit = Signal(str)
    # Signal to emit a move action in form of a tuple (param_name, value)
    move_action = Signal(tuple)

    def __init__(
        self,
        parent=None,
        client=None,
        config=None,
        target_widget=None,
        gui_id: str | None = None,
        ui_file="lmfit_dialog_vertical.ui",
        **kwargs,
    ):
        """
        Initialises the LMFitDialog widget.

        Args:
            parent (QWidget): The parent widget.
            client: BEC client object.
            config: Configuration of the widget.
            target_widget: The widget that the settings will be taken from and applied to.
            gui_id (str): GUI ID.
            ui_file (str): The UI file to be loaded.
        """
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)
        self.setProperty("skip_settings", True)
        self._ui_file = ui_file
        self.target_widget = target_widget

        current_path = os.path.dirname(__file__)
        self.ui = UILoader(self).loader(os.path.join(current_path, self._ui_file))
        self.layout = QVBoxLayout(self)
        self.layout.addWidget(self.ui)
        self.summary_data = {}
        self._fit_curve_id = None
        self._deci_precision = 3
        self._always_show_latest = False
        self.ui.curve_list.currentItemChanged.connect(self.display_fit_details)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        self._active_actions = []
        self._enable_actions = True
        self._move_buttons = []
        self._accent_colors = get_accent_colors()
        self.action_buttons = {}

    @property
    def enable_actions(self) -> bool:
        """SafeProperty to enable the move to buttons."""
        return self._enable_actions

    @enable_actions.setter
    def enable_actions(self, enable: bool):
        self._enable_actions = enable
        for button in self.action_buttons.values():
            button.setEnabled(enable)

    @SafeProperty(list)
    def active_action_list(self) -> list[str]:
        """SafeProperty to list the names of the fit parameters for which actions should be enabled."""
        return self._active_actions

    @active_action_list.setter
    def active_action_list(self, actions: list[str]):
        self._active_actions = actions

    # This SafeSlot needed?
    @SafeSlot(bool)
    def set_actions_enabled(self, enable: bool) -> bool:
        """SafeSlot to enable the move to buttons.

        Args:
            enable (bool): Whether to enable the action buttons.
        """
        self.enable_actions = enable

    @SafeProperty(bool)
    def always_show_latest(self):
        """SafeProperty to indicate if always the latest DAP update is displayed."""
        return self._always_show_latest

    @always_show_latest.setter
    def always_show_latest(self, show: bool):
        self._always_show_latest = show

    @SafeProperty(bool)
    def hide_curve_selection(self):
        """SafeProperty for showing the curve selection."""
        return not self.ui.group_curve_selection.isVisible()

    @hide_curve_selection.setter
    def hide_curve_selection(self, show: bool):
        """Setter for showing the curve selection.

        Args:
            show (bool): Whether to show the curve selection.
        """
        self.ui.group_curve_selection.setVisible(not show)

    @SafeProperty(bool)
    def hide_summary(self) -> bool:
        """SafeProperty for showing the summary."""
        return not self.ui.group_summary.isVisible()

    @hide_summary.setter
    def hide_summary(self, show: bool):
        """Setter for showing the summary.

        Args:
            show (bool): Whether to show the summary.
        """
        self.ui.group_summary.setVisible(not show)

    @SafeProperty(bool)
    def hide_parameters(self) -> bool:
        """SafeProperty for showing the parameters."""
        return not self.ui.group_parameters.isVisible()

    @hide_parameters.setter
    def hide_parameters(self, show: bool):
        """Setter for showing the parameters.

        Args:
            show (bool): Whether to show the parameters.
        """
        self.ui.group_parameters.setVisible(not show)

    @property
    def fit_curve_id(self) -> str:
        """SafeProperty for the currently displayed fit curve_id."""
        return self._fit_curve_id

    @fit_curve_id.setter
    def fit_curve_id(self, curve_id: str):
        """Setter for the currently displayed fit curve_id.

        Args:
            fit_curve_id (str): The curve_id of the fit curve to be displayed.
        """
        self._fit_curve_id = curve_id
        self.selected_fit.emit(curve_id)

    @SafeSlot(str)
    def remove_dap_data(self, curve_id: str):
        """Remove the DAP data for the given curve_id.

        Args:
            curve_id (str): The curve_id of the DAP data to be removed.
        """
        self.summary_data.pop(curve_id, None)
        self.refresh_curve_list()

    @SafeSlot(str)
    def select_curve(self, curve_id: str):
        """Select active curve_id in the curve list.

        Args:
            curve_id (str): curve_id to be selected.
        """
        self.fit_curve_id = curve_id

    @SafeSlot(dict, dict)
    def update_summary_tree(self, data: dict, metadata: dict):
        """Update the summary tree with the given data.

        Args:
            data (dict): Data for the DAP Summary.
            metadata (dict): Metadata of the fit curve.
        """
        curve_id = metadata.get("curve_id", "")
        self.summary_data.update({curve_id: data})
        self.refresh_curve_list()
        if self.fit_curve_id is None or self.always_show_latest is True:
            self.fit_curve_id = curve_id
        if curve_id != self.fit_curve_id:
            return
        if data is None:
            return
        self.ui.summary_tree.clear()
        chi_squared = data.get("chisqr", 0.0)
        if isinstance(chi_squared, float) or isinstance(chi_squared, int):
            chi_squared = f"{chi_squared:.{self._deci_precision}f}"
        else:
            chi_squared = "None"
        reduced_chi_squared = data.get("redchi", 0.0)
        if isinstance(reduced_chi_squared, float) or isinstance(reduced_chi_squared, int):
            reduced_chi_squared = f"{reduced_chi_squared:.{self._deci_precision}f}"
        else:
            reduced_chi_squared = "None"
        r_squared = data.get("rsquared", 0.0)
        if isinstance(r_squared, float) or isinstance(r_squared, int):
            r_squared = f"{r_squared:.{self._deci_precision}f}"
        else:
            r_squared = "None"
        properties = [
            ("Model", data.get("model", "")),
            ("Method", data.get("method", "")),
            ("Chi-Squared", chi_squared),
            ("Reduced Chi-Squared", reduced_chi_squared),
            ("R-Squared", r_squared),
            ("Message", data.get("message", "")),
        ]
        for prop, val in properties:
            QTreeWidgetItem(self.ui.summary_tree, [prop, val])
        self.update_param_tree(data.get("params", []))

    def _update_summary_data(self, curve_id: str, data: dict):
        """Update the summary data with the given data.

        Args:
            curve_id (str): The curve_id of the fit curve.
            data (dict): The data to be updated.
        """
        self.summary_data.update({curve_id: data})
        if self.fit_curve_id is not None:
            return
        self.fit_curve_id = curve_id

    def update_param_tree(self, params):
        """Update the parameter tree with the given parameters.

        Args:
            params (list): List of LMFit parameters for the fit curve.
        """
        self._move_buttons = []
        self.ui.param_tree.clear()
        for param in params:
            param_name = param[0]
            param_value = param[1]
            if isinstance(param_value, float) or isinstance(param_value, int):
                param_value = f"{param_value:.{self._deci_precision}f}"
            else:
                param_value = "None"
            param_std = param[7]
            if isinstance(param_std, float) or isinstance(param_std, int):
                param_std = f"{param_std:.{self._deci_precision}f}"
            else:
                param_std = "None"

            tree_item = QTreeWidgetItem(self.ui.param_tree, [param_name, param_value, param_std])
            if param_name in self.active_action_list:  # pylint: disable=unsupported-membership-test
                # Create a push button to move the motor to a specific position
                widget = QWidget()
                button = QPushButton(f"Move to {param_name}")
                button.clicked.connect(self._create_move_action(param_name, param[1]))
                if self.enable_actions is True:
                    button.setEnabled(True)
                else:
                    button.setEnabled(False)
                button.setStyleSheet(
                    f"""
                    QPushButton:enabled {{ background-color: {self._accent_colors.success.name()};color: white; }} 
                    QPushButton:disabled {{ background-color: grey;color: white; }}
                    """
                )
                self.action_buttons[param_name] = button
                layout = QVBoxLayout()
                layout.addWidget(self.action_buttons[param_name])
                layout.setContentsMargins(0, 0, 0, 0)
                widget.setLayout(layout)
                self.ui.param_tree.setItemWidget(tree_item, 3, widget)

    def _create_move_action(self, param_name: str, param_value: float) -> callable:
        """Create a move action for the given parameter name and value.

        Args:
            param_name (str): The name of the parameter.
            param_value (float): The value of the parameter.
        Returns:
            callable: The move action with the given parameter name and value.
        """

        def move_action():
            self.move_action.emit((param_name, param_value))

        return move_action

    def populate_curve_list(self):
        """Populate the curve list with the available fit curves."""
        for curve_name in self.summary_data:
            self.ui.curve_list.addItem(curve_name)

    def refresh_curve_list(self):
        """Refresh the curve list with the updated data."""
        self.ui.curve_list.clear()
        self.populate_curve_list()

    def display_fit_details(self, current):
        """Callback for displaying the fit details of the selected curve.

        Args:
            current: The current item in the curve list.
        """
        if current:
            curve_name = current.text()
            self.fit_curve_id = curve_name
            data = self.summary_data[curve_name]
            if data is None:
                return
            self.update_summary_tree(data, {"curve_id": curve_name})


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication  # pylint: disable=ungrouped-imports

    app = QApplication(sys.argv)
    dialog = LMFitDialog()
    dialog.show()
    sys.exit(app.exec_())
